"""Admin endpoints for activations and integration usage."""

from __future__ import annotations

import logging
from datetime import datetime, timezone
from typing import Any, Dict, List

from fastapi import APIRouter, Depends, Query
from sqlalchemy import text
from sqlalchemy.ext.asyncio import AsyncSession

from ...auth_security import AdminContext, require_admin
from ...db import get_session
from ...models import WorkspaceSecret, Integration

log = logging.getLogger(__name__)

router = APIRouter(tags=["Admin Activations"])


def _now_iso() -> str:
    return datetime.now(timezone.utc).isoformat()


def _dt_iso(value: Any) -> Any:
    if value is None:
        return None
    if isinstance(value, datetime):
        return value.isoformat()
    return str(value)


@router.get("/overview")
async def get_activations_overview(
    admin_ctx: AdminContext = Depends(require_admin),
    db: AsyncSession = Depends(get_session),
) -> Dict[str, Any]:
    workspace_table = WorkspaceSecret.__tablename__
    integration_table = Integration.__tablename__

    byo_query = text(
        f"""
        SELECT integration,
               COUNT(*) AS connections,
               COUNT(DISTINCT workspace_id) AS workspaces,
               COUNT(*) FILTER (WHERE validated_at IS NOT NULL) AS validated,
               MAX(created_at) AS most_recent
        FROM {workspace_table}
        GROUP BY integration
        ORDER BY connections DESC
        """
    )
    oauth_query = text(
        f"""
        SELECT provider,
               COUNT(*) AS connections,
               COUNT(DISTINCT account_id) AS accounts,
               COUNT(*) FILTER (WHERE status IN ('connected','active')) AS active,
               MAX(created_at) AS most_recent
        FROM {integration_table}
        GROUP BY provider
        ORDER BY connections DESC
        """
    )

    byo_totals_query = text(
        f"""
        SELECT COUNT(DISTINCT workspace_id) AS workspaces,
               COUNT(*) AS connections,
               COUNT(*) FILTER (WHERE validated_at IS NOT NULL) AS validated
        FROM {workspace_table}
        """
    )
    oauth_totals_query = text(
        f"""
        SELECT COUNT(DISTINCT account_id) AS accounts,
               COUNT(*) AS connections,
               COUNT(*) FILTER (WHERE status IN ('connected','active')) AS active
        FROM {integration_table}
        """
    )

    byo_rows = (await db.execute(byo_query)).mappings().all()
    oauth_rows = (await db.execute(oauth_query)).mappings().all()
    byo_totals = (await db.execute(byo_totals_query)).mappings().one_or_none() or {}
    oauth_totals = (await db.execute(oauth_totals_query)).mappings().one_or_none() or {}

    byo_providers = []
    for row in byo_rows:
        connections = row.get("connections") or 0
        validated = row.get("validated") or 0
        byo_providers.append(
            {
                "integration": row.get("integration"),
                "workspaces": row.get("workspaces") or 0,
                "connections": connections,
                "validated": validated,
                "most_recent": _dt_iso(row.get("most_recent")),
                "validation_rate": (validated / connections) if connections else 0,
            }
        )

    oauth_providers = []
    for row in oauth_rows:
        connections = row.get("connections") or 0
        active = row.get("active") or 0
        oauth_providers.append(
            {
                "provider": row.get("provider"),
                "accounts": row.get("accounts") or 0,
                "connections": connections,
                "active": active,
                "most_recent": _dt_iso(row.get("most_recent")),
                "activation_rate": (active / connections) if connections else 0,
            }
        )

    total_connections = (byo_totals.get("connections") or 0) + (
        oauth_totals.get("connections") or 0
    )

    summary = {
        "byo_providers": len(byo_providers),
        "byo_workspaces": byo_totals.get("workspaces") or 0,
        "byo_connections": byo_totals.get("connections") or 0,
        "byo_validated": byo_totals.get("validated") or 0,
        "oauth_providers": len(oauth_providers),
        "oauth_accounts": oauth_totals.get("accounts") or 0,
        "oauth_connections": oauth_totals.get("connections") or 0,
        "oauth_active": oauth_totals.get("active") or 0,
        "total_connections": total_connections,
        "generated_at": _now_iso(),
    }

    return {
        "summary": summary,
        "byo_providers": byo_providers,
        "oauth_providers": oauth_providers,
    }


@router.get("/recent-connections")
async def get_recent_connections(
    days: int = Query(30, ge=1, le=365),
    limit: int = Query(50, ge=1, le=200),
    admin_ctx: AdminContext = Depends(require_admin),
    db: AsyncSession = Depends(get_session),
) -> Dict[str, Any]:
    workspace_table = WorkspaceSecret.__tablename__
    integration_table = Integration.__tablename__

    byo_recent_query = text(
        f"""
        SELECT integration, workspace_id, key_name, created_at, updated_at, validated_at
        FROM {workspace_table}
        WHERE created_at >= (now() - (:days * interval '1 day'))
        ORDER BY created_at DESC
        LIMIT :limit
        """
    )
    oauth_recent_query = text(
        f"""
        SELECT provider, account_id, status, created_at, updated_at
        FROM {integration_table}
        WHERE created_at >= (now() - (:days * interval '1 day'))
        ORDER BY created_at DESC
        LIMIT :limit
        """
    )

    params = {"days": days, "limit": limit}
    byo_rows = (await db.execute(byo_recent_query, params)).mappings().all()
    oauth_rows = (await db.execute(oauth_recent_query, params)).mappings().all()

    items: List[Dict[str, Any]] = []
    for row in byo_rows:
        items.append(
            {
                "source": "byo",
                "provider": row.get("integration"),
                "workspace_id": row.get("workspace_id"),
                "account_id": None,
                "status": None,
                "key_name": row.get("key_name"),
                "validated_at": _dt_iso(row.get("validated_at")),
                "created_at": _dt_iso(row.get("created_at")),
                "updated_at": _dt_iso(row.get("updated_at")),
            }
        )
    for row in oauth_rows:
        items.append(
            {
                "source": "oauth",
                "provider": row.get("provider"),
                "workspace_id": None,
                "account_id": row.get("account_id"),
                "status": row.get("status"),
                "key_name": None,
                "validated_at": None,
                "created_at": _dt_iso(row.get("created_at")),
                "updated_at": _dt_iso(row.get("updated_at")),
            }
        )

    _epoch = datetime.min.replace(tzinfo=timezone.utc)

    def _sort_key(item: Dict[str, Any]):
        value = item.get("created_at")
        if isinstance(value, str):
            try:
                dt = datetime.fromisoformat(value.replace("Z", "+00:00"))
                if dt.tzinfo is None:
                    dt = dt.replace(tzinfo=timezone.utc)
                return dt
            except ValueError:
                return _epoch
        if isinstance(value, datetime):
            if value.tzinfo is None:
                return value.replace(tzinfo=timezone.utc)
            return value
        return _epoch

    items.sort(key=_sort_key, reverse=True)
    if len(items) > limit:
        items = items[:limit]

    return {
        "items": items,
        "days": days,
        "limit": limit,
        "generated_at": _now_iso(),
    }


@router.get("/usage-by-provider")
async def get_usage_by_provider(
    days: int = Query(30, ge=1, le=365),
    admin_ctx: AdminContext = Depends(require_admin),
    db: AsyncSession = Depends(get_session),
) -> Dict[str, Any]:
    totals_query = text(
        """
        SELECT source,
               COUNT(*) AS events,
               COUNT(DISTINCT org_id) AS unique_orgs,
               COALESCE(SUM(rows), 0) AS total_rows
        FROM saas_events
        WHERE ts >= (now() - (:days * interval '1 day'))
        GROUP BY source
        ORDER BY events DESC
        """
    )
    actions_query = text(
        """
        SELECT source, action, COUNT(*) AS events
        FROM saas_events
        WHERE ts >= (now() - (:days * interval '1 day'))
        GROUP BY source, action
        ORDER BY source, events DESC
        """
    )

    params = {"days": days}
    totals_rows = (await db.execute(totals_query, params)).mappings().all()
    actions_rows = (await db.execute(actions_query, params)).mappings().all()

    actions_by_source: Dict[str, List[Dict[str, Any]]] = {}
    for row in actions_rows:
        source = row.get("source") or "unknown"
        actions_by_source.setdefault(source, []).append(
            {"action": row.get("action"), "events": row.get("events") or 0}
        )

    sources: List[Dict[str, Any]] = []
    for row in totals_rows:
        source = row.get("source") or "unknown"
        actions = actions_by_source.get(source, [])
        actions_sorted = sorted(actions, key=lambda x: x["events"], reverse=True)
        sources.append(
            {
                "source": source,
                "events": row.get("events") or 0,
                "unique_orgs": row.get("unique_orgs") or 0,
                "total_rows": row.get("total_rows") or 0,
                "top_actions": actions_sorted[:6],
            }
        )

    return {
        "sources": sources,
        "days": days,
        "generated_at": _now_iso(),
    }


@router.get("/validation-status")
async def get_validation_status(
    admin_ctx: AdminContext = Depends(require_admin),
    db: AsyncSession = Depends(get_session),
) -> Dict[str, Any]:
    workspace_table = WorkspaceSecret.__tablename__
    stale_days = 30

    status_query = text(
        f"""
        SELECT integration,
               COUNT(*) AS total,
               COUNT(*) FILTER (
                    WHERE validated_at IS NOT NULL
                      AND validated_at >= (now() - (:stale_days * interval '1 day'))
               ) AS validated,
               COUNT(*) FILTER (WHERE validated_at IS NULL) AS unvalidated,
               COUNT(*) FILTER (
                    WHERE validated_at IS NOT NULL
                      AND validated_at < (now() - (:stale_days * interval '1 day'))
               ) AS stale,
               MAX(created_at) AS most_recent
        FROM {workspace_table}
        GROUP BY integration
        ORDER BY total DESC
        """
    )

    rows = (await db.execute(status_query, {"stale_days": stale_days})).mappings().all()
    providers: List[Dict[str, Any]] = []
    for row in rows:
        total = row.get("total") or 0
        validated = row.get("validated") or 0
        providers.append(
            {
                "integration": row.get("integration"),
                "total": total,
                "validated": validated,
                "unvalidated": row.get("unvalidated") or 0,
                "stale": row.get("stale") or 0,
                "most_recent": _dt_iso(row.get("most_recent")),
                "validation_rate": (validated / total) if total else 0,
            }
        )

    return {
        "providers": providers,
        "stale_days": stale_days,
        "generated_at": _now_iso(),
    }



"""Security utilities for Gremlin Autopilot.

Provides:
- File path validation and allowlisting
- URL validation against webhook profiles
- LLM context consent enforcement
- Parameter sanitization
"""

from __future__ import annotations

import fnmatch
import re
from pathlib import Path

from .models import (
    ConsentError,
    LLMContext,
    LLMContextLevel,
    SampleOverrideToken,
    SheetContext,
    UserSettings,
    ValidationError,
)

# Shell metacharacters that could be dangerous
SHELL_METACHARACTERS = re.compile(r'[;&|`$(){}[\]<>\\"\'\n\r\x00]')


def get_default_allowed_paths() -> list[Path]:
    """Get the default list of allowed paths.

    Only ~/.g_gremlin/ is allowed by default. CWD is NOT included.
    """
    return [Path.home() / ".g_gremlin"]


def get_allowed_paths(settings: UserSettings) -> list[Path]:
    """Get allowed paths including user-configured ones.

    Args:
        settings: User settings containing allowed_paths

    Returns:
        List of allowed Path objects
    """
    paths = get_default_allowed_paths()

    for path_str in settings.allowed_paths:
        try:
            path = Path(path_str).expanduser().resolve()
            paths.append(path)
        except (OSError, ValueError):
            # Skip invalid paths
            continue

    return paths


def validate_file_path(path: str, allowed_roots: list[Path]) -> Path:
    """Validate a file path against the allowlist.

    Args:
        path: The path to validate (must be absolute)
        allowed_roots: List of allowed root directories

    Returns:
        The resolved Path if valid

    Raises:
        ValidationError: If path is invalid or outside allowed directories
    """
    # Reject relative paths
    if not path.startswith(("/", "~")) and ":" not in path:  # Allow Windows absolute paths
        raise ValidationError(f"Absolute path required, got: {path}")

    try:
        # Expand user and resolve to absolute path
        resolved = Path(path).expanduser().resolve()
    except (OSError, ValueError) as e:
        raise ValidationError(f"Invalid path: {e}")

    # Check if path is under any allowed root
    for root in allowed_roots:
        try:
            resolved.relative_to(root)
            return resolved
        except ValueError:
            continue

    raise ValidationError(
        f"Path {path} is outside allowed directories. "
        f"Allowed: {[str(r) for r in allowed_roots]}"
    )


def validate_webhook_url(url: str, settings: UserSettings) -> bool:
    """Validate a webhook URL against configured profiles.

    Args:
        url: The URL to validate
        settings: User settings containing webhook_profiles

    Returns:
        True if URL matches a profile

    Raises:
        ValidationError: If URL doesn't match any profile
    """
    if not settings.webhook_profiles:
        raise ValidationError(
            "No webhook profiles configured. "
            "Add profiles in settings before using webhook URLs."
        )

    for profile in settings.webhook_profiles:
        pattern = profile.get("url", "")
        if fnmatch.fnmatch(url, pattern):
            return True

    raise ValidationError(
        f"URL {url} is not in allowed webhook profiles. "
        f"Allowed patterns: {[p.get('url') for p in settings.webhook_profiles]}"
    )


def sanitize_tool_arg(value: str) -> str:
    """Sanitize a tool argument by removing shell metacharacters.

    Args:
        value: The string value to sanitize

    Returns:
        Sanitized string with metacharacters removed
    """
    return SHELL_METACHARACTERS.sub("", value)


def validate_tool_args(tool_name: str, args: dict) -> dict:
    """Validate tool arguments for safety.

    Args:
        tool_name: Name of the tool
        args: Dictionary of arguments

    Returns:
        Validated arguments

    Raises:
        ValidationError: If any argument contains dangerous content
    """
    validated = {}

    for key, value in args.items():
        if isinstance(value, str):
            # Check for shell metacharacters
            if SHELL_METACHARACTERS.search(value):
                raise ValidationError(
                    f"Tool '{tool_name}' argument '{key}' contains "
                    f"invalid characters: {value!r}"
                )
            validated[key] = value
        elif isinstance(value, list):
            # Validate list items
            validated[key] = [
                validate_tool_args(tool_name, {"item": v})["item"]
                if isinstance(v, str)
                else v
                for v in value
            ]
        else:
            validated[key] = value

    return validated


def build_llm_context(
    sheet_context: SheetContext,
    settings: UserSettings,
) -> LLMContext:
    """Build LLM context with consent enforcement.

    Args:
        sheet_context: The current sheet context
        settings: User settings including consent

    Returns:
        LLMContext with appropriate data based on consent

    Raises:
        ConsentError: If samples are present but consent not given
    """
    # Check consent for samples
    if sheet_context.sample_rows:
        if settings.llm_context_level != LLMContextLevel.WITH_SAMPLES:
            raise ConsentError(
                "Sample rows require explicit consent in settings. "
                "Enable 'with_samples' in Data Sharing Preferences."
            )
        if not settings.consent_given_at:
            raise ConsentError(
                "Consent timestamp missing for sample sharing. "
                "Please re-confirm Data Sharing Preferences."
            )

    # Build context based on consent level
    sample_rows = None
    if settings.llm_context_level == LLMContextLevel.WITH_SAMPLES:
        sample_rows = sheet_context.sample_rows

    return LLMContext(
        sheet_id=sheet_context.sheet_id,
        tab_name=sheet_context.tab_name,
        headers=sheet_context.headers,
        row_count=sheet_context.row_count,
        column_types=sheet_context.column_types,
        sample_rows=sample_rows,
    )


def build_llm_context_with_override(
    sheet_context: SheetContext,
    settings: UserSettings,
    override_token: SampleOverrideToken | None,
) -> LLMContext:
    """Build LLM context with optional per-request override.

    Args:
        sheet_context: The current sheet context
        settings: User settings
        override_token: Optional override token for temporary samples

    Returns:
        LLMContext with appropriate data

    Raises:
        ConsentError: If neither settings nor valid override permits samples
    """
    # If override token is valid, allow samples
    if override_token and override_token.is_valid():
        return LLMContext(
            sheet_id=sheet_context.sheet_id,
            tab_name=sheet_context.tab_name,
            headers=sheet_context.headers,
            row_count=sheet_context.row_count,
            column_types=sheet_context.column_types,
            sample_rows=sheet_context.sample_rows[:override_token.sample_count]
            if sheet_context.sample_rows
            else None,
        )

    # Fall back to normal consent check
    return build_llm_context(sheet_context, settings)



"""
Dynamics 365 Audit Bridge Module
Integrates with FoundryMatch audit system
"""

import logging
from typing import Dict, Any, Optional
from datetime import datetime
import json

logger = logging.getLogger(__name__)


class DynamicsAuditIntegration:
    """
    Bridge between Dynamics integration and FoundryMatch audit system
    Mirrors SalesforceAuditIntegration interface
    """

    def __init__(self, audit_logger, integration):
        """
        Initialize audit bridge

        Args:
            audit_logger: Main AuditLogger instance
            integration: DynamicsIntegration instance
        """
        self.audit_logger = audit_logger
        self.integration = integration
        self.session_id = None

    async def start_session(self, metadata: Dict[str, Any] = None) -> str:
        """Start new audit session"""
        session_metadata = {
            "integration": "dynamics365",
            "user": self.integration.credentials.client_id,
            "resource": self.integration.credentials.resource,
            "start_time": datetime.utcnow().isoformat()
        }

        if metadata:
            session_metadata.update(metadata)

        self.session_id = await self.audit_logger.start_session(
            session_type="dynamics_integration",
            metadata=session_metadata
        )

        return self.session_id

    async def end_session(self, summary: Dict[str, Any] = None):
        """End current audit session"""
        if not self.session_id:
            return

        session_summary = {
            "end_time": datetime.utcnow().isoformat(),
            "api_calls": getattr(self, "_api_call_count", 0),
            "errors": getattr(self, "_error_count", 0)
        }

        if summary:
            session_summary.update(summary)

        await self.audit_logger.end_session(self.session_id, session_summary)
        self.session_id = None

    async def log_api_call(self, method: str, endpoint: str,
                          status_code: int, duration_ms: float,
                          error: Optional[str] = None):
        """Log API call details"""
        # Track call count
        if not hasattr(self, "_api_call_count"):
            self._api_call_count = 0
        self._api_call_count += 1

        # Track errors
        if error or status_code >= 400:
            if not hasattr(self, "_error_count"):
                self._error_count = 0
            self._error_count += 1

        event_data = {
            "method": method,
            "endpoint": endpoint,
            "status_code": status_code,
            "duration_ms": duration_ms,
            "timestamp": datetime.utcnow().isoformat()
        }

        if error:
            event_data["error"] = error

        # Add rate limit info if available
        if hasattr(self.integration, "rate_limiter"):
            rate_limit_status = self.integration.rate_limiter.get_status()
            event_data["rate_limit"] = {
                "burst_remaining": rate_limit_status["burst_tokens"],
                "service_remaining": rate_limit_status["service_tokens"]
            }

        await self.audit_logger.log_event(
            event_type="dynamics_api_call",
            event_data=event_data,
            session_id=self.session_id
        )

    async def log_authentication(self, success: bool, details: Dict[str, Any]):
        """Log authentication attempts"""
        event_data = {
            "success": success,
            "timestamp": datetime.utcnow().isoformat(),
            "method": details.get("method", "oauth"),
            "client_id": details.get("client_id"),
            "tenant_id": details.get("tenant_id")
        }

        if "error" in details:
            event_data["error"] = details["error"]

        if "duration_ms" in details:
            event_data["duration_ms"] = details["duration_ms"]

        await self.audit_logger.log_event(
            event_type="dynamics_authentication",
            event_data=event_data,
            session_id=self.session_id
        )

    async def log_batch_operation(self, operation: str, total_records: int,
                                 successful: int, failed: int,
                                 duration_ms: float):
        """Log batch operation results"""
        event_data = {
            "operation": operation,
            "total_records": total_records,
            "successful": successful,
            "failed": failed,
            "duration_ms": duration_ms,
            "success_rate": (successful / total_records) if total_records > 0 else 0,
            "timestamp": datetime.utcnow().isoformat()
        }

        await self.audit_logger.log_event(
            event_type="dynamics_batch_operation",
            event_data=event_data,
            session_id=self.session_id
        )

    async def log_error(self, error_type: str, error_details: Any,
                       context: Optional[Dict[str, Any]] = None):
        """Log error details"""
        event_data = {
            "error_type": error_type,
            "timestamp": datetime.utcnow().isoformat()
        }

        # Handle different error detail types
        if isinstance(error_details, Exception):
            event_data["error_message"] = str(error_details)
            event_data["error_class"] = error_details.__class__.__name__
        elif isinstance(error_details, dict):
            event_data.update(error_details)
        else:
            event_data["error_details"] = str(error_details)

        if context:
            event_data["context"] = context

        await self.audit_logger.log_event(
            event_type="dynamics_error",
            event_data=event_data,
            session_id=self.session_id
        )

    async def log_event(self, event_name: str, event_data: Dict[str, Any]):
        """Log generic event"""
        event_data["timestamp"] = datetime.utcnow().isoformat()

        await self.audit_logger.log_event(
            event_type=f"dynamics_{event_name}",
            event_data=event_data,
            session_id=self.session_id
        )

    async def log_data_validation(self, entity: str, total_rows: int,
                                 valid_rows: int, invalid_rows: int,
                                 validation_errors: list):
        """Log data validation results"""
        event_data = {
            "entity": entity,
            "total_rows": total_rows,
            "valid_rows": valid_rows,
            "invalid_rows": invalid_rows,
            "validation_rate": (valid_rows / total_rows) if total_rows > 0 else 0,
            "error_count": len(validation_errors),
            "timestamp": datetime.utcnow().isoformat()
        }

        # Include sample of errors (first 10)
        if validation_errors:
            event_data["sample_errors"] = [
                {
                    "row": err.row_index,
                    "field": err.field,
                    "error_type": err.error_type,
                    "message": err.message
                }
                for err in validation_errors[:10]
            ]

        await self.audit_logger.log_event(
            event_type="dynamics_data_validation",
            event_data=event_data,
            session_id=self.session_id
        )

    async def log_workflow_execution(self, workflow_name: str,
                                   params: Dict[str, Any],
                                   result: Any):
        """Log workflow execution details"""
        event_data = {
            "workflow_name": workflow_name,
            "timestamp": datetime.utcnow().isoformat(),
            "params": self._sanitize_params(params),
            "status": result.status,
            "total_processed": result.total_processed,
            "successful": result.successful,
            "failed": result.failed,
            "execution_time": result.execution_time
        }

        if result.errors:
            event_data["error_count"] = len(result.errors)
            event_data["sample_errors"] = result.errors[:5]

        await self.audit_logger.log_event(
            event_type="dynamics_workflow_execution",
            event_data=event_data,
            session_id=self.session_id
        )

    def _sanitize_params(self, params: Dict[str, Any]) -> Dict[str, Any]:
        """Sanitize parameters for logging (remove sensitive data)"""
        sanitized = {}

        for key, value in params.items():
            if key in ["password", "client_secret", "access_token", "refresh_token"]:
                sanitized[key] = "***REDACTED***"
            elif isinstance(value, list) and len(value) > 100:
                # Truncate large lists
                sanitized[key] = f"<list of {len(value)} items>"
            elif isinstance(value, dict):
                # Recursively sanitize nested dicts
                sanitized[key] = self._sanitize_params(value)
            else:
                sanitized[key] = value

        return sanitized

    async def get_session_summary(self) -> Dict[str, Any]:
        """Get summary of current session"""
        if not self.session_id:
            return {}

        summary = await self.audit_logger.get_session_summary(self.session_id)

        # Add Dynamics-specific metrics
        if summary:
            events = summary.get("events", [])

            # Calculate API call metrics
            api_calls = [e for e in events if e["event_type"] == "dynamics_api_call"]
            if api_calls:
                total_duration = sum(e["event_data"].get("duration_ms", 0) for e in api_calls)
                summary["api_metrics"] = {
                    "total_calls": len(api_calls),
                    "total_duration_ms": total_duration,
                    "avg_duration_ms": total_duration / len(api_calls),
                    "error_count": sum(1 for e in api_calls if e["event_data"].get("error"))
                }

            # Calculate batch operation metrics
            batch_ops = [e for e in events if e["event_type"] == "dynamics_batch_operation"]
            if batch_ops:
                summary["batch_metrics"] = {
                    "total_operations": len(batch_ops),
                    "total_records": sum(e["event_data"]["total_records"] for e in batch_ops),
                    "total_successful": sum(e["event_data"]["successful"] for e in batch_ops),
                    "total_failed": sum(e["event_data"]["failed"] for e in batch_ops)
                }

        return summary



"""People Data Labs API endpoints for Google Sheets integration"""

import hashlib
import json
from typing import Any, Dict, List, Optional
from fastapi import APIRouter, HTTPException, Depends, Request
from pydantic import BaseModel, Field
from sqlalchemy.ext.asyncio import AsyncSession

from fmatch.saas.integrations.people_data_labs import (
    PeopleDataLabsAdapter,
    PDLCompanyEnrichmentRequest,
    PDLPersonEnrichmentRequest,
)
from fmatch.saas.auth import verify_api_key
from fmatch.saas.db import get_session
from fmatch.saas.config.enrichment import ENRICHMENT_ENABLED
from fmatch.saas.services.enrichment_usage_tracker import EnrichmentUsageTracker


def require_enrichment_enabled():
    if not ENRICHMENT_ENABLED:
        raise HTTPException(
            status_code=404,
            detail="Enrichment features are not currently available",
        )


router = APIRouter(
    prefix="/api/v1/pdl",
    tags=["pdl"],
    dependencies=[Depends(require_enrichment_enabled)],
)

ENRICH_FIELD_PREFIX = "enrich_"


class PDLEnrichRequest(BaseModel):
    """Request for PDL enrichment from Google Sheets"""

    source: Dict[str, Any] = Field(..., description="Source data with headers and rows")
    requests: List[Dict[str, Any]] = Field(..., description="PDL enrichment requests")
    provider: str = Field(default="pdl")
    type: str = Field(..., description="Enrichment type: company or person")
    output_fields: List[str] = Field(
        default_factory=list, description="Fields to include in output"
    )


class PDLSearchRequest(BaseModel):
    """Request for PDL search"""

    type: str = Field(..., description="Search type: company_search or person_search")
    sql_query: Optional[str] = Field(None, description="SQL query for search")
    es_query: Optional[Dict] = Field(None, description="Elasticsearch query for search")
    max_results: int = Field(default=100, ge=1, le=1000)
    output_fields: List[str] = Field(default_factory=list)
    provider: str = Field(default="pdl")


class PDLEnrichResponse(BaseModel):
    """Response for PDL enrichment"""

    ok: bool = True
    headers: List[str]
    rows: List[List[Any]]
    matched: int = 0
    credits: float = 0
    errors: List[str] = Field(default_factory=list)


class PDLBalanceResponse(BaseModel):
    """Balance snapshot for enrichment credits"""

    account_id: str
    credits_available: int = 0
    allowance_remaining: int = 0
    pack_remaining: int = 0
    daily_remaining: Optional[int] = None
    monthly_remaining: Optional[int] = None
    credits_per_request: float = 0.0
    price_per_request_usd: float = 0.0
    day_start_utc: Optional[str] = None
    next_day_start_utc: Optional[str] = None
    is_free_tier: bool = False
    raw: Dict[str, Any] = Field(default_factory=dict)


class PDLEstimateResponse(BaseModel):
    """Estimate for enrichment credit usage"""

    estimated_requests: int = 0


class PDLJobsRequest(BaseModel):
    """Request to fetch job postings"""

    company_name: Optional[str] = None
    company_domain: Optional[str] = None
    job_title: Optional[str] = None
    location: Optional[str] = None
    size: int = Field(default=10, ge=1, le=100)


class PDLTableResponse(BaseModel):
    ok: bool = True
    headers: List[str]
    rows: List[List[Any]]
    matched: int = 0
    credits: float = 0
    errors: List[str] = Field(default_factory=list)


class PDLTechExtractionRequest(BaseModel):
    """Company technographics extraction (batch)"""

    requests: List[Dict[str, Any]]
    min_likelihood: int = Field(default=0, ge=0, le=10)


class PDLEmploymentExtractionRequest(BaseModel):
    """Employment linkage extraction (batch)"""

    requests: List[Dict[str, Any]]
    min_likelihood: int = Field(default=0, ge=0, le=10)


@router.post("/enrich", response_model=PDLEnrichResponse)
async def enrich_data(
    request: PDLEnrichRequest,
    raw_request: Request,
    api_key: str = Depends(verify_api_key),
    db: AsyncSession = Depends(get_session),
) -> PDLEnrichResponse:
    """Batch enrichment endpoint for Google Sheets with usage tracking"""

    adapter = PeopleDataLabsAdapter()

    if not adapter.available():
        raise HTTPException(status_code=503, detail="PDL API key not configured")

    # Initialize usage tracker
    tracker = EnrichmentUsageTracker(db)
    account_id, user_email = tracker.resolve_account(raw_request)

    # Generate idempotency key and request hash for tracking
    payload = json.dumps(
        {
            "type": request.type,
            "requests": request.requests[:100],
            "output_fields": sorted(request.output_fields),
        },
        sort_keys=True,
    )
    payload_hash = hashlib.sha256(payload.encode()).hexdigest()
    idem_key = f"pdl-sheets-{request.type}-{payload_hash[:16]}"

    source_headers = request.source.get("headers", [])
    source_rows = request.source.get("rows", [])

    if not source_rows:
        return PDLEnrichResponse(
            ok=True, headers=source_headers, rows=[], matched=0, credits=0
        )

    # Estimate and check quota
    estimated_requests = min(len(request.requests), 100)
    preflight = await tracker.preflight(
        account_id=account_id,
        estimated_requests=estimated_requests,
        idempotency_key=idem_key,
        request_hash=payload_hash,
        reserve=True,
        job_id=f"sheets-enrich-{request.type}",
    )

    # Handle idempotent replay
    if preflight.get("replay"):
        raise HTTPException(
            status_code=409,
            detail="This request was already processed. Check your previous results.",
        )

    # Handle quota denial
    if not preflight.get("allowed"):
        raise HTTPException(
            status_code=preflight.get("status_code", 429),
            detail={
                "error": "quota_exceeded",
                "message": preflight.get("reason", "Insufficient credits"),
                "allowance_remaining": preflight.get("allowance_remaining", 0),
                "pack_remaining": preflight.get("pack_remaining", 0),
            },
        )

    try:
        pdl_requests = _build_pdl_requests(request)
        if request.type == "company":
            result = adapter.bulk_company_enrichment(pdl_requests)
        else:
            result = adapter.bulk_person_enrichment(pdl_requests)

        actual_requests = len(pdl_requests)

        if "error" in result:
            # Release reservation on error
            await tracker.release(account_id, idem_key)
            raise HTTPException(status_code=500, detail=result["error"])

        # Process responses and build output rows
        responses = result.get("responses", [])
        enriched_headers = list(source_headers)
        enriched_rows = []
        matched = 0

        # Determine which enrichment fields to include
        enrich_fields_to_add = set()
        for field in request.output_fields:
            column_name = f"{ENRICH_FIELD_PREFIX}{field}"
            if column_name not in enriched_headers:
                enriched_headers.append(column_name)
                enrich_fields_to_add.add(field)

        # Process each row
        for i, source_row in enumerate(source_rows):
            enriched_row = list(source_row)

            # Add enrichment data if available
            if i < len(responses) and responses[i]:
                resp = responses[i]
                if resp.get("status") == 200 and "data" in resp:
                    matched += 1
                    data = resp["data"]

                    # Add requested fields
                    for field in enrich_fields_to_add:
                        value = _extract_field_value(data, field)
                        enriched_row.append(value)
                else:
                    # No match - add empty values
                    enriched_row.extend([""] * len(enrich_fields_to_add))
            else:
                # No response - add empty values
                enriched_row.extend([""] * len(enrich_fields_to_add))

            enriched_rows.append(enriched_row)

        # Finalize usage tracking
        await tracker.finalize(
            account_id=account_id,
            user_email=user_email,
            idempotency_key=idem_key,
            request_hash=payload_hash,
            actual_requests=actual_requests,  # Provider charges per request, not match
            matches_found=matched,
            operation="enrich",
            enrichment_type=request.type,
            provider="pdl",
            job_id=f"sheets-enrich-{request.type}",
            chunk_id=None,
            meta={
                "records_scanned": len(source_rows),
                "candidates_found": len(request.requests),
                "source": "google_sheets",
                "output_fields": request.output_fields,
            },
        )

        return PDLEnrichResponse(
            ok=True,
            headers=enriched_headers,
            rows=enriched_rows,
            matched=matched,
            credits=actual_requests,  # Accurate credit count
        )

    except HTTPException:
        raise
    except Exception as e:
        # Try to release reservation on unexpected error
        try:
            await tracker.release(account_id, idem_key)
        except Exception:
            pass
        raise HTTPException(status_code=500, detail=str(e))


@router.get("/balances", response_model=PDLBalanceResponse)
async def get_pdl_balances(
    raw_request: Request,
    api_key: str = Depends(verify_api_key),
    db: AsyncSession = Depends(get_session),
) -> PDLBalanceResponse:
    """Expose current enrichment credit balances for the requesting account."""

    tracker = EnrichmentUsageTracker(db)
    account_id, _ = tracker.resolve_account(raw_request)

    balances = await tracker.balances(account_id)

    allowance_remaining = int(balances.get("allowance_remaining") or 0)
    pack_remaining = int(balances.get("pack_remaining") or 0)
    credits_available = allowance_remaining + pack_remaining
    price_per_request = float(balances.get("price_per_request_usd") or 0.0)

    def _optional_int(value: Any) -> Optional[int]:
        if value is None:
            return None
        try:
            return int(value)
        except (TypeError, ValueError):
            return None

    # Treat zero-priced tiers or empty balances as free tier for UX messaging.
    is_free_tier = price_per_request <= 0 or credits_available <= 0

    return PDLBalanceResponse(
        account_id=str(account_id),
        credits_available=credits_available,
        allowance_remaining=allowance_remaining,
        pack_remaining=pack_remaining,
        daily_remaining=_optional_int(balances.get("daily_remaining")),
        monthly_remaining=_optional_int(balances.get("monthly_remaining")),
        credits_per_request=float(balances.get("credits_per_request") or 0.0),
        price_per_request_usd=price_per_request,
        day_start_utc=balances.get("day_start_utc"),
        next_day_start_utc=balances.get("next_day_start_utc"),
        is_free_tier=is_free_tier,
        raw=balances,
    )


@router.post("/estimate", response_model=PDLEstimateResponse)
async def estimate_enrichment(
    request: PDLEnrichRequest, api_key: str = Depends(verify_api_key)
) -> PDLEstimateResponse:
    """Return the number of provider requests required for this enrichment payload."""

    adapter = PeopleDataLabsAdapter()
    if not adapter.available():
        raise HTTPException(status_code=503, detail="PDL API key not configured")

    pdl_requests = _build_pdl_requests(request)
    return PDLEstimateResponse(estimated_requests=len(pdl_requests))


@router.post("/search", response_model=PDLEnrichResponse)
async def search_data(
    request: PDLSearchRequest, api_key: str = Depends(verify_api_key)
) -> PDLEnrichResponse:
    """Search endpoint for Google Sheets"""

    adapter = PeopleDataLabsAdapter()

    if not adapter.available():
        raise HTTPException(status_code=503, detail="PDL API key not configured")

    try:
        # Perform search based on type
        if request.type == "company_search":
            result = adapter.search_companies(
                sql_query=request.sql_query,
                es_query=request.es_query,
                size=request.max_results,
            )
        else:  # person_search
            result = adapter.search_people(
                sql_query=request.sql_query,
                es_query=request.es_query,
                size=request.max_results,
            )

        if "error" in result:
            raise HTTPException(status_code=500, detail=result["error"])

        # Extract data from search results
        data_list = result.get("data", [])

        if not data_list:
            return PDLEnrichResponse(ok=True, headers=[], rows=[], matched=0, credits=0)

        # Build headers based on requested fields or all available fields
        if request.output_fields:
            headers = request.output_fields
        else:
            # Use all fields from first result
            headers = list(data_list[0].keys()) if data_list else []

        # Build rows
        rows = []
        for item in data_list:
            row = []
            for field in headers:
                value = _extract_field_value(item, field)
                row.append(value)
            rows.append(row)

        return PDLEnrichResponse(
            ok=True,
            headers=headers,
            rows=rows,
            matched=len(rows),
            credits=len(rows),  # Search API charges per result
        )

    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


def _build_pdl_requests(payload: PDLEnrichRequest) -> List[Any]:
    """Convert incoming payload into provider request objects."""

    trimmed_requests = payload.requests[:100]
    requests: List[Any] = []

    if payload.type == "company":
        for req_dict in trimmed_requests:
            requests.append(
                PDLCompanyEnrichmentRequest(
                    name=req_dict.get("name"),
                    website=req_dict.get("website"),
                    domain=req_dict.get("domain"),
                    ticker=req_dict.get("ticker"),
                    linkedin_url=req_dict.get("linkedin_url"),
                    location=req_dict.get("location"),
                    min_likelihood=req_dict.get("min_likelihood", 0),
                )
            )
    elif payload.type == "person":
        for req_dict in trimmed_requests:
            requests.append(
                PDLPersonEnrichmentRequest(
                    email=req_dict.get("email"),
                    first_name=req_dict.get("first_name"),
                    last_name=req_dict.get("last_name"),
                    company=req_dict.get("company"),
                    phone=req_dict.get("phone"),
                    profile=req_dict.get("profile") or req_dict.get("linkedin_url"),
                    min_likelihood=req_dict.get("min_likelihood", 0),
                )
            )
    else:
        raise HTTPException(
            status_code=400,
            detail=f"Unsupported enrichment type '{payload.type}'. Use 'company' or 'person'.",
        )

    return requests


def _extract_field_value(data: Dict, field: str) -> Any:
    """Extract a field value from PDL data, handling nested structures"""

    # Handle nested field access (e.g., "job.title")
    if "." in field:
        parts = field.split(".")
        value = data
        for part in parts:
            if isinstance(value, dict):
                value = value.get(part)
            else:
                return ""
        return _format_value(value)

    # Handle special fields
    if field == "work_email":
        # Try to get work email from various sources
        if "work_email" in data:
            return data["work_email"]
        elif "emails" in data and isinstance(data["emails"], list) and data["emails"]:
            # Return first email marked as current/primary
            for email in data["emails"]:
                if isinstance(email, dict) and email.get("current"):
                    return email.get("address", "")
            # Fallback to first email
            return (
                data["emails"][0].get("address", "")
                if isinstance(data["emails"][0], dict)
                else data["emails"][0]
            )
        return ""

    elif field == "phone_numbers":
        if "phone_numbers" in data and isinstance(data["phone_numbers"], list):
            # Return comma-separated list of phone numbers
            phones = []
            for phone in data["phone_numbers"]:
                if isinstance(phone, dict):
                    phones.append(phone.get("number", ""))
                else:
                    phones.append(str(phone))
            return ", ".join(filter(None, phones))
        return ""

    elif field == "personal_emails":
        # Gather non-work emails if available
        emails = data.get("emails")
        out = []
        if isinstance(emails, list):
            for e in emails:
                if isinstance(e, dict):
                    addr = e.get("address") or e.get("email")
                    typ = (e.get("type") or "").lower()
                    if addr and (typ in ("personal", "other") or not typ):
                        out.append(addr)
                elif isinstance(e, str):
                    out.append(e)
        return ", ".join(dict.fromkeys(out)) if out else ""

    elif field == "linkedin_url":
        # Try various common places
        if data.get("linkedin_url"):
            return data.get("linkedin_url")
        profiles = data.get("profiles") or data.get("social_profiles")
        if isinstance(profiles, list):
            for p in profiles:
                if isinstance(p, dict) and (
                    p.get("site") == "linkedin"
                    or "linkedin.com" in str(p.get("url") or "")
                ):
                    return p.get("url", "")
        return ""

    elif field == "technologies":
        if "technologies" in data and isinstance(data["technologies"], list):
            # Return comma-separated list of technologies
            return ", ".join(data["technologies"])
        return ""

    elif field == "job_title":
        # Try to get current job title
        if "job_title" in data:
            return data["job_title"]
        elif (
            "experience" in data
            and isinstance(data["experience"], list)
            and data["experience"]
        ):
            # Get title from most recent/current experience
            for exp in data["experience"]:
                if isinstance(exp, dict) and exp.get("current"):
                    return exp.get("title", "")
            # Fallback to first experience
            return (
                data["experience"][0].get("title", "")
                if isinstance(data["experience"][0], dict)
                else ""
            )
        return ""

    elif field == "job_company_name":
        # Try to get current company name
        if "job_company_name" in data:
            return data["job_company_name"]
        elif (
            "experience" in data
            and isinstance(data["experience"], list)
            and data["experience"]
        ):
            for exp in data["experience"]:
                if isinstance(exp, dict) and exp.get("current"):
                    company = exp.get("company", {})
                    if isinstance(company, dict):
                        return company.get("name", "")
                    return str(company)
            # Fallback to first experience
            if isinstance(data["experience"][0], dict):
                company = data["experience"][0].get("company", {})
                if isinstance(company, dict):
                    return company.get("name", "")
        return ""

    elif field == "job_company_website":
        # Try to get current company website
        if "job_company_website" in data:
            return data["job_company_website"]
        exp = data.get("experience")
        if isinstance(exp, list) and exp:
            for item in exp:
                if isinstance(item, dict) and item.get("current"):
                    company = item.get("company")
                    if isinstance(company, dict):
                        return company.get("website", "")
        return ""

    # Default field extraction
    value = data.get(field, "")
    return _format_value(value)


def _format_value(value: Any) -> str:
    """Format a value for spreadsheet output"""

    if value is None:
        return ""
    elif isinstance(value, bool):
        return "Yes" if value else "No"
    elif isinstance(value, (list, dict)):
        # Convert complex types to string representation
        if isinstance(value, list) and value and isinstance(value[0], str):
            # Simple string list - join with commas
            return ", ".join(value)
        # For complex structures, use string representation
        return str(value)
    else:
        return str(value)


@router.post("/jobs", response_model=PDLTableResponse)
async def get_job_postings(
    request: PDLJobsRequest,
    api_key: str = Depends(verify_api_key),
) -> PDLTableResponse:
    """Return recent job postings as a table for account intelligence."""
    adapter = PeopleDataLabsAdapter()
    if not adapter.available():
        raise HTTPException(status_code=503, detail="PDL API key not configured")

    result = adapter.get_job_postings(
        company_name=request.company_name,
        company_domain=request.company_domain,
        job_title=request.job_title,
        location=request.location,
        size=request.size,
    )
    if "error" in result:
        raise HTTPException(status_code=500, detail=result["error"])
    data_list = result.get("data", [])
    if not data_list:
        return PDLTableResponse(ok=True, headers=[], rows=[], matched=0, credits=0)

    # Choose common job fields
    headers = [
        "job_title",
        "company_name",
        "company_domain",
        "location",
        "posted_at",
        "url",
    ]
    rows: List[List[Any]] = []
    for item in data_list:
        row = [
            _extract_field_value(item, "job_title"),
            _extract_field_value(item, "company_name"),
            _extract_field_value(item, "company_domain"),
            _extract_field_value(item, "location"),
            _extract_field_value(item, "posted_at"),
            _extract_field_value(item, "url"),
        ]
        rows.append(row)

    return PDLTableResponse(
        ok=True, headers=headers, rows=rows, matched=len(rows), credits=float(len(rows))
    )


@router.post("/tech", response_model=PDLTableResponse)
async def extract_technographics(
    request: PDLTechExtractionRequest,
    api_key: str = Depends(verify_api_key),
) -> PDLTableResponse:
    """Extract company technologies for a batch of company queries."""
    adapter = PeopleDataLabsAdapter()
    if not adapter.available():
        raise HTTPException(status_code=503, detail="PDL API key not configured")

    # Build PDLCompanyEnrichmentRequest list
    pdl_requests: List[PDLCompanyEnrichmentRequest] = []
    for req in request.requests[:100]:
        pdl_requests.append(
            PDLCompanyEnrichmentRequest(
                name=req.get("name"),
                website=req.get("website"),
                domain=req.get("domain"),
                linkedin_url=req.get("linkedin_url"),
                min_likelihood=req.get("min_likelihood", request.min_likelihood),
            )
        )

    result = adapter.bulk_company_enrichment(pdl_requests)
    if "error" in result:
        raise HTTPException(status_code=500, detail=result["error"])

    responses = result.get("responses", [])
    headers = ["name", "domain", "technologies"]
    rows: List[List[Any]] = []

    for resp in responses:
        if resp and resp.get("status") in (200, None) and "data" in resp:
            data = resp["data"]
            name = _extract_field_value(data, "name")
            domain = _extract_field_value(data, "domain") or _extract_field_value(
                data, "website"
            )
            techs = _extract_field_value(data, "technologies")
            rows.append([name, domain, techs])
        else:
            rows.append(["", "", ""])  # preserve alignment

    return PDLTableResponse(
        ok=True, headers=headers, rows=rows, matched=len([r for r in rows if any(r)])
    )


@router.post("/employment", response_model=PDLTableResponse)
async def extract_employment(
    request: PDLEmploymentExtractionRequest,
    api_key: str = Depends(verify_api_key),
) -> PDLTableResponse:
    """Extract current (and basic past) employment info for people.

    Returns a table with: full_name, job_title, job_company_name, job_company_website,
    job_start_date, job_end_date, job_is_current
    """
    adapter = PeopleDataLabsAdapter()
    if not adapter.available():
        raise HTTPException(status_code=503, detail="PDL API key not configured")

    # Build person enrichment requests
    pdl_requests: List[PDLPersonEnrichmentRequest] = []
    for req in request.requests[:100]:
        pdl_requests.append(
            PDLPersonEnrichmentRequest(
                email=req.get("email"),
                first_name=req.get("first_name"),
                last_name=req.get("last_name"),
                company=req.get("company"),
                profile=req.get("profile") or req.get("linkedin_url"),
                min_likelihood=req.get("min_likelihood", request.min_likelihood),
            )
        )

    result = adapter.bulk_person_enrichment(pdl_requests)
    if "error" in result:
        raise HTTPException(status_code=500, detail=result["error"])

    responses = result.get("responses", [])
    headers = [
        "full_name",
        "job_title",
        "job_company_name",
        "job_company_website",
        "job_start_date",
        "job_end_date",
        "job_is_current",
    ]
    rows: List[List[Any]] = []

    for resp in responses:
        if resp and resp.get("status") in (200, None) and "data" in resp:
            data = resp["data"]
            # full_name with fallback
            full_name = data.get("full_name") or (
                (
                    (data.get("first_name") or "") + " " + (data.get("last_name") or "")
                ).strip()
            )
            # current employment from experience
            title = _extract_field_value(data, "job_title")
            company = _extract_field_value(data, "job_company_name")
            website = _extract_field_value(data, "job_company_website")
            start = ""
            end = ""
            is_current = ""
            exp = data.get("experience")
            if isinstance(exp, list) and exp:
                cur = None
                for e in exp:
                    if isinstance(e, dict) and e.get("current"):
                        cur = e
                        break
                if not cur and isinstance(exp[0], dict):
                    cur = exp[0]
                if isinstance(cur, dict):
                    start = cur.get("start_date") or cur.get("start") or ""
                    end = cur.get("end_date") or cur.get("end") or ""
                    is_current = bool(cur.get("current"))
            rows.append([full_name, title, company, website, start, end, is_current])
        else:
            rows.append(["", "", "", "", "", "", ""])  # preserve alignment

    return PDLTableResponse(
        ok=True, headers=headers, rows=rows, matched=len([r for r in rows if any(r)])
    )



"""Thin client for interacting with Google Gemini via Vertex AI."""

from __future__ import annotations

import concurrent.futures
import json
import logging
import os
import random
import re
import time
from typing import Any, Dict, List, Optional

import vertexai
from vertexai.generative_models import GenerativeModel, GenerationConfig

from .llm_guard import (
    LLMConcurrencyError,
    LLMCircuitOpenError,
    LLMGuard,
    record_failure,
    record_success,
)

# Import Google API exceptions for retry handling
try:
    from google.api_core import exceptions as google_exceptions
    HAS_GOOGLE_EXCEPTIONS = True
except ImportError:
    HAS_GOOGLE_EXCEPTIONS = False

logger = logging.getLogger(__name__)

# Retry configuration
MAX_RETRIES = 3
INITIAL_BACKOFF_S = 1.0
MAX_BACKOFF_S = 16.0
BACKOFF_MULTIPLIER = 2.0
_EXECUTOR = concurrent.futures.ThreadPoolExecutor(max_workers=8)

class GeminiClient:
    """Wrapper around Gemini via Vertex AI."""

    def __init__(
        self,
        model: str,
        timeout_s: int = 12,
        project_id: Optional[str] = None,
        location: str = "us-central1",
        max_output_tokens: Optional[int] = None,
        max_retries: Optional[int] = None,
        wall_clock_budget_s: Optional[int] = None,
        retry_jitter_min_ms: Optional[int] = None,
        retry_jitter_max_ms: Optional[int] = None,
        account_id: Optional[str] = None,
    ) -> None:
        """
        Initialize Gemini client using Vertex AI.

        Args:
            model: Gemini model name (e.g., "gemini-2.0-flash-exp")
            timeout_s: Request timeout in seconds
            project_id: GCP project ID (uses default from gcloud if None)
            location: GCP region
        """
        self.model_name = model
        self.timeout = timeout_s
        self.max_retries = (
            max_retries if max_retries is not None else max(0, MAX_RETRIES - 1)
        )
        self.wall_clock_budget_s = wall_clock_budget_s
        self.retry_jitter_min_ms = retry_jitter_min_ms
        self.retry_jitter_max_ms = retry_jitter_max_ms
        self.account_id = account_id
        if (model or "").strip().lower().startswith("gemini-3-"):
            location = "global"

        # Get project ID from env or gcloud config
        if project_id is None:
            project_id = os.getenv("GOOGLE_CLOUD_PROJECT") or "aqueous-encoder-478015-r7"

        self.project_id = project_id
        self.location = location

        # Initialize Vertex AI (uses gcloud auth)
        vertexai.init(project=project_id, location=location)

        if max_output_tokens is None:
            env_max = os.getenv("GEMINI_MAX_OUTPUT_TOKENS")
            if env_max:
                try:
                    max_output_tokens = int(env_max)
                except ValueError:
                    max_output_tokens = 2048
            else:
                max_output_tokens = 2048

        # Initialize model with JSON response config
        generation_config = GenerationConfig(
            temperature=0.1,  # Low for consistency
            top_p=0.9,
            top_k=40,
            max_output_tokens=max_output_tokens,
            response_mime_type="application/json",  # Force JSON output
        )

        self.model = GenerativeModel(
            model_name=model,
            generation_config=generation_config,
        )

    def _is_retryable_error(self, exc: Exception) -> bool:
        """Check if an exception is retryable (rate limit, service unavailable, etc.)."""
        if isinstance(exc, TimeoutError):
            return True
        if HAS_GOOGLE_EXCEPTIONS:
            # 429 Resource Exhausted (rate limit)
            if isinstance(exc, google_exceptions.ResourceExhausted):
                return True
            # 503 Service Unavailable
            if isinstance(exc, google_exceptions.ServiceUnavailable):
                return True
            # 500 Internal Server Error
            if isinstance(exc, google_exceptions.InternalServerError):
                return True
            # 504 Gateway Timeout
            if isinstance(exc, google_exceptions.DeadlineExceeded):
                return True

        # Fallback: check error message for common retry scenarios
        err_str = str(exc).lower()
        retryable_patterns = [
            "rate limit",
            "resource exhausted",
            "quota exceeded",
            "service unavailable",
            "503",
            "429",
            "timeout",
            "deadline exceeded",
        ]
        return any(pattern in err_str for pattern in retryable_patterns)

    def _sanitize_error(self, exc: Exception) -> str:
        """Sanitize error message to avoid exposing sensitive information."""
        err_str = str(exc)
        # Remove potential credential/token patterns
        err_str = re.sub(r"(Bearer\s+)[^\s]+", r"\1[REDACTED]", err_str)
        err_str = re.sub(r"(api[_-]?key[=:]\s*)[^\s&]+", r"\1[REDACTED]", err_str, flags=re.IGNORECASE)
        err_str = re.sub(r"(token[=:]\s*)[^\s&]+", r"\1[REDACTED]", err_str, flags=re.IGNORECASE)
        # Truncate long messages
        if len(err_str) > 500:
            err_str = err_str[:500] + "..."
        return err_str

    def _call_with_timeout(self, content: str, timeout_s: Optional[float]) -> Any:
        if timeout_s is None or timeout_s <= 0:
            return self.model.generate_content(content)
        future = _EXECUTOR.submit(self.model.generate_content, content)
        try:
            return future.result(timeout=timeout_s)
        except concurrent.futures.TimeoutError as exc:
            future.cancel()
            raise TimeoutError("Gemini request timed out") from exc

    def _retry_sleep_s(self, backoff: float) -> float:
        if self.retry_jitter_min_ms is not None and self.retry_jitter_max_ms is not None:
            low = min(self.retry_jitter_min_ms, self.retry_jitter_max_ms)
            high = max(self.retry_jitter_min_ms, self.retry_jitter_max_ms)
            return random.uniform(low, high) / 1000.0
        return backoff

    def _generate_with_retry(self, content: str, account_id: Optional[str]) -> Any:
        """Call generate_content with retry logic for transient errors."""
        last_exc: Optional[Exception] = None
        backoff = INITIAL_BACKOFF_S
        start = time.monotonic()
        total_attempts = self.max_retries + 1
        budget_s = (
            float(self.wall_clock_budget_s)
            if self.wall_clock_budget_s is not None
            else float(self.timeout) * total_attempts
        )

        for attempt in range(total_attempts):
            elapsed = time.monotonic() - start
            remaining = budget_s - elapsed
            if remaining <= 0:
                break
            try:
                timeout_s = min(self.timeout, remaining)
                with LLMGuard(
                    model=self.model_name,
                    account_id=account_id,
                    timeout_s=timeout_s,
                ):
                    response = self._call_with_timeout(content, timeout_s)
                record_success(self.model_name)
                return response
            except (LLMConcurrencyError, LLMCircuitOpenError):
                raise
            except Exception as exc:
                last_exc = exc
                retryable = self._is_retryable_error(exc)
                if not retryable:
                    raise
                if attempt >= self.max_retries:
                    record_failure(self.model_name)
                    raise
                sleep_s = self._retry_sleep_s(backoff)
                logger.warning(
                    "Gemini API error (attempt %d/%d), retrying in %.2fs: %s",
                    attempt + 1,
                    total_attempts,
                    sleep_s,
                    type(exc).__name__,
                )
                time.sleep(sleep_s)
                if self.retry_jitter_min_ms is None:
                    backoff = min(backoff * BACKOFF_MULTIPLIER, MAX_BACKOFF_S)

        # All retries exhausted
        if last_exc:
            record_failure(self.model_name)
            raise last_exc
        raise RuntimeError("Unexpected retry loop exit")

    def chat_json(
        self,
        messages: List[Dict[str, Any]],
        retry_strict: bool = True,
        account_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Call Gemini's chat API expecting a JSON response payload.

        Args:
            messages: List of {"role": "user"|"system", "content": str}
            retry_strict: If True, retry once with stricter prompt on parse failure

        Returns:
            Parsed JSON dict from model response

        Raises:
            ValueError: If response is not valid JSON after retries
            Exception: On API errors
        """
        # Convert messages to Gemini format
        gemini_messages = self._convert_messages(messages)

        try:
            # Generate response with retry logic for transient errors
            response = self._generate_with_retry(
                gemini_messages, account_id or self.account_id
            )

            # Extract text
            content_text = response.text.strip()

            if not content_text:
                return {"_parse_error": True, "_raw": "(empty response)"}

            # Parse JSON
            try:
                return self._parse_json_response(content_text)
            except (json.JSONDecodeError, ValueError) as parse_err:
                logger.warning(
                    "Gemini JSON parse failed (first pass). Raw=%s",
                    content_text[:2000],
                )
                # Retry with stricter prompt if enabled
                if retry_strict:
                    return self._retry_with_strict_prompt(messages, content_text)
                else:
                    return {
                        "_parse_error": True,
                        "_raw": content_text,
                        "_error": str(parse_err)
                    }

        except Exception as api_err:
            # Log and return sanitized error
            sanitized = self._sanitize_error(api_err)
            logger.error("Gemini API error after retries: %s", sanitized)
            err_type = type(api_err).__name__
            retry_after = None
            if isinstance(api_err, (LLMConcurrencyError, LLMCircuitOpenError)):
                retry_after = getattr(api_err, "retry_after_s", None)
            return {
                "_parse_error": True,
                "_raw": sanitized,
                "_error": f"API error: {err_type}",
                "_error_type": err_type,
                "_retry_after_s": retry_after,
            }

    def _convert_messages(
        self, messages: List[Dict[str, Any]]
    ) -> str:
        """
        Convert OpenAI-style messages to Gemini format (simple string for Vertex AI).

        Gemini doesn't have a "system" role, so we prepend system messages
        to the first user message.
        """
        system_parts = []
        user_parts = []

        for msg in messages:
            role = msg.get("role", "user")
            content = msg.get("content", "")

            if role == "system":
                system_parts.append(content)
            else:
                user_parts.append(content)

        # Combine system + user messages into a single string
        if system_parts:
            full_content = "\n\n".join(system_parts) + "\n\n" + "\n\n".join(user_parts)
        else:
            full_content = "\n\n".join(user_parts)

        return full_content

    def _parse_json_response(self, content: str) -> Dict[str, Any]:
        """
        Parse JSON from response, handling common LLM quirks.

        - Strips markdown code fences (```json ... ```)
        - Handles trailing commas
        - Cleans control characters
        """
        # Remove invalid control characters that break JSON parsing
        content = re.sub(r"[\x00-\x08\x0b\x0c\x0e-\x1f]", "", content)

        # Normalize smart quotes to straight quotes
        content = (
            content.replace("\u201c", '"')
            .replace("\u201d", '"')
            .replace("\u2018", "'")
            .replace("\u2019", "'")
        )

        # Remove markdown code fences
        content = re.sub(r"^```(?:json)?\s*", "", content, flags=re.MULTILINE)
        content = re.sub(r"\s*```\s*$", "", content, flags=re.MULTILINE)
        content = content.strip()

        # Try to find JSON object/array bounds
        if not content.startswith(("{", "[")):
            # Search for first { or [
            match = re.search(r"[{\[]", content)
            if match:
                content = content[match.start():]

        # Trim anything after the final closing brace/bracket
        end_idx = max(content.rfind("}"), content.rfind("]"))
        if end_idx != -1:
            content = content[: end_idx + 1]

        # Extract the first balanced JSON object/array if multiple are present
        content = self._extract_first_json(content)

        # Escape unescaped control chars/newlines inside JSON strings
        content = self._escape_control_chars_in_strings(content)
        # Remove stray lines that are just a quote (common LLM glitch)
        content = re.sub(r'^\s*"\s*$', "", content, flags=re.MULTILINE)

        # Parse
        try:
            return json.loads(content)
        except json.JSONDecodeError as parse_err:
            logger.debug(
                "Gemini JSON decode error: %s at line %s col %s",
                parse_err.msg,
                getattr(parse_err, "lineno", "?"),
                getattr(parse_err, "colno", "?"),
            )
            # Try fixing trailing commas (common LLM error)
            fixed = re.sub(r",\s*([}\]])", r"\1", content)
            try:
                return json.loads(fixed)
            except json.JSONDecodeError as parse_err2:
                logger.debug(
                    "Gemini JSON decode error after comma fix: %s at line %s col %s",
                    parse_err2.msg,
                    getattr(parse_err2, "lineno", "?"),
                    getattr(parse_err2, "colno", "?"),
                )
                # As a last resort, try parsing with strict=False (if available)
                try:
                    return json.loads(fixed, strict=False)  # type: ignore[arg-type]
                except Exception:
                    logger.debug("Gemini JSON parse failed: %s", fixed[:200])
                    raise

    @staticmethod
    def _extract_first_json(content: str) -> str:
        """Return the first balanced JSON object/array from content."""
        start_idx = None
        depth = 0
        in_str = False
        escape = False
        for i, ch in enumerate(content):
            if start_idx is None:
                if ch in "{[":
                    start_idx = i
                    depth = 1
                continue
            if escape:
                escape = False
                continue
            if ch == "\\":
                escape = True
                continue
            if ch == '"':
                in_str = not in_str
                continue
            if in_str:
                continue
            if ch in "{[":
                depth += 1
            elif ch in "}]":
                depth -= 1
                if depth == 0:
                    return content[start_idx : i + 1]
        return content

    @staticmethod
    def _escape_control_chars_in_strings(content: str) -> str:
        """Escape newlines/tabs within JSON string literals."""
        out = []
        in_str = False
        escape = False
        for ch in content:
            if escape:
                out.append(ch)
                escape = False
                continue
            if ch == "\\":
                out.append(ch)
                escape = True
                continue
            if ch == '"':
                in_str = not in_str
                out.append(ch)
                continue
            if in_str:
                if ch == "\n":
                    out.append("\\n")
                    continue
                if ch == "\r":
                    out.append("\\r")
                    continue
                if ch == "\t":
                    out.append("\\t")
                    continue
            out.append(ch)
        return "".join(out)

    def _retry_with_strict_prompt(
        self,
        original_messages: List[Dict[str, Any]],
        failed_response: str
    ) -> Dict[str, Any]:
        """Retry with a stricter JSON prompt."""
        strict_suffix = (
            "\n\nIMPORTANT: Respond with ONLY valid JSON. "
            "No markdown, no explanation, no code fences. "
            "Just pure JSON starting with { or [. "
            "Output MUST be minified on a single line (no newlines). "
            "Do NOT include unescaped quotes or newlines inside string values."
        )

        # Append strict instruction to last message
        retry_messages = original_messages.copy()
        if retry_messages:
            retry_messages[-1]["content"] += strict_suffix

        try:
            gemini_messages = self._convert_messages(retry_messages)
            # Use retry wrapper for transient errors
            response = self._generate_with_retry(gemini_messages)
            content_text = response.text.strip()
            return self._parse_json_response(content_text)
        except Exception as retry_err:
            sanitized = self._sanitize_error(retry_err)
            logger.warning(
                "Gemini JSON parse failed after retry. Raw=%s",
                failed_response[:2000],
            )
            return {
                "_parse_error": True,
                "_raw": failed_response,
                "_retry_raw": sanitized,
                "_error": "JSON parse failed after retry"
            }



"""Render Scout findings for CLI output."""

from __future__ import annotations

from typing import Any, Iterable, List

from .models import ScoutFinding, ScoutScanResult


def _truncate(value: str, max_len: int = 64) -> str:
    if len(value) <= max_len:
        return value
    return value[: max_len - 3] + "..."


def _format_value(value: Any) -> str:
    if value is None:
        return ""
    text = str(value)
    return _truncate(text)


def _format_table(headers: List[str], rows: List[List[str]]) -> str:
    widths = [len(h) for h in headers]
    for row in rows:
        for idx, cell in enumerate(row):
            widths[idx] = max(widths[idx], len(cell))
    sep = "+" + "+".join("-" * (w + 2) for w in widths) + "+"
    header_line = "|" + "|".join(f" {headers[i].ljust(widths[i])} " for i in range(len(headers))) + "|"
    lines = [sep, header_line, sep]
    for row in rows:
        line = "|" + "|".join(f" {row[i].ljust(widths[i])} " for i in range(len(headers))) + "|"
        lines.append(line)
    lines.append(sep)
    return "\n".join(lines)


def _summarize_fields(fields: dict[str, Any], max_items: int = 3) -> str:
    if not fields:
        return ""
    items: List[str] = []
    for idx, (key, value) in enumerate(fields.items()):
        if idx >= max_items:
            break
        items.append(f"{key}={_format_value(value)}")
    return "; ".join(items)


def render_finding_table(finding: ScoutFinding, include_fields: bool = True) -> str:
    title = finding.display_name or finding.scout_name
    lines: List[str] = []
    lines.append(f"Scout: {title} ({finding.scout_name})")
    lines.append(
        f"Status: {finding.status} | Total scanned: {finding.total_records_scanned} | "
        f"Findings: {finding.findings_count} | Duration: {finding.duration_ms} ms"
    )
    if finding.category or finding.severity:
        lines.append(f"Category: {finding.category or ''} | Severity: {finding.severity or ''}")
    if finding.status_detail:
        lines.append(f"Detail: {finding.status_detail}")
    if finding.warnings:
        lines.append("Warnings: " + "; ".join(finding.warnings))

    if not finding.evidence:
        return "\n".join(lines)

    headers = ["Record Id", "Reason Code", "Reason", "SObject"]
    if include_fields:
        headers.append("Fields")

    rows: List[List[str]] = []
    for row in finding.evidence:
        row_values = [
            _format_value(row.record_id),
            _format_value(row.reason_code),
            _format_value(row.reason),
            _format_value(row.sobject),
        ]
        if include_fields:
            row_values.append(_summarize_fields(row.fields))
        rows.append(row_values)

    lines.append("")
    lines.append(_format_table(headers, rows))
    return "\n".join(lines)


def render_scan_summary(scan_result: ScoutScanResult) -> str:
    headers = ["Scout", "Status", "Findings", "Total", "Duration (ms)"]
    rows: List[List[str]] = []
    for finding in scan_result.findings:
        title = finding.display_name or finding.scout_name
        rows.append(
            [
                _format_value(title),
                _format_value(finding.status),
                _format_value(str(finding.findings_count)),
                _format_value(str(finding.total_records_scanned)),
                _format_value(str(finding.duration_ms)),
            ]
        )
    summary = _format_table(headers, rows)
    meta = (
        f"Scout scan {scan_result.run_id} | Started: {scan_result.started_at} | "
        f"Ended: {scan_result.ended_at} | Duration: {scan_result.total_duration_ms} ms"
    )
    return "\n".join([meta, summary])


def render_scan_with_findings(scan_result: ScoutScanResult, include_fields: bool = True) -> str:
    parts = [render_scan_summary(scan_result)]
    for finding in scan_result.findings:
        parts.append("")
        parts.append(render_finding_table(finding, include_fields=include_fields))
    return "\n".join(parts)


__all__ = ["render_finding_table", "render_scan_summary", "render_scan_with_findings"]



"""SQLite-backed suppression store for Autopilot snooze."""

from __future__ import annotations

import hashlib
import sqlite3
from contextlib import contextmanager
from dataclasses import replace
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Iterable, List

from g_gremlin.scout.models import EvidenceRow, ScoutFinding


def _utc_now() -> datetime:
    return datetime.now(timezone.utc)


def _utc_now_iso() -> str:
    return _utc_now().isoformat()


def _parse_iso(value: str | None) -> datetime | None:
    if not value:
        return None
    try:
        return datetime.fromisoformat(value.replace("Z", "+00:00"))
    except Exception:
        return None


def _finding_hash(scout_name: str, record_id: str, reason_code: str) -> str:
    base = f"{scout_name}:{record_id}:{reason_code}"
    return hashlib.sha256(base.encode("utf-8")).hexdigest()


def _chunks(items: List[str], size: int = 500) -> Iterable[List[str]]:
    for i in range(0, len(items), size):
        yield items[i : i + size]


class SuppressStore:
    def __init__(self, db_path: Path) -> None:
        self.db_path = Path(db_path)
        self.db_path.parent.mkdir(parents=True, exist_ok=True)
        self._ensure_schema()

    @contextmanager
    def _connect(self) -> Iterable[sqlite3.Connection]:
        conn = sqlite3.connect(self.db_path)
        conn.row_factory = sqlite3.Row
        try:
            yield conn
            conn.commit()
        except Exception:
            conn.rollback()
            raise
        finally:
            conn.close()

    def _ensure_schema(self) -> None:
        with self._connect() as conn:
            conn.execute(
                """
                CREATE TABLE IF NOT EXISTS suppressed_findings (
                    finding_hash TEXT PRIMARY KEY,
                    scout_name TEXT,
                    record_id TEXT,
                    reason_code TEXT,
                    created_at TEXT,
                    expires_at TEXT,
                    comment TEXT
                )
                """
            )
            conn.execute(
                """
                CREATE TABLE IF NOT EXISTS suppressed_runs (
                    run_id TEXT PRIMARY KEY,
                    created_at TEXT,
                    expires_at TEXT,
                    comment TEXT
                )
                """
            )

    def suppress_findings(
        self,
        findings: Iterable[ScoutFinding],
        *,
        days: int = 7,
        comment: str = "",
    ) -> int:
        expires_at = _utc_now() + timedelta(days=days)
        now = _utc_now_iso()
        count = 0
        with self._connect() as conn:
            for finding in findings:
                for row in finding.evidence:
                    if not row.record_id:
                        continue
                    hash_value = _finding_hash(finding.scout_name, row.record_id, row.reason_code)
                    conn.execute(
                        """
                        INSERT INTO suppressed_findings (
                            finding_hash, scout_name, record_id, reason_code,
                            created_at, expires_at, comment
                        )
                        VALUES (?, ?, ?, ?, ?, ?, ?)
                        ON CONFLICT(finding_hash) DO UPDATE SET
                          expires_at = excluded.expires_at,
                          comment = excluded.comment
                        """,
                        (
                            hash_value,
                            finding.scout_name,
                            row.record_id,
                            row.reason_code,
                            now,
                            expires_at.isoformat(),
                            comment,
                        ),
                    )
                    count += 1
        return count

    def suppress_run(self, run_id: str, *, days: int = 7, comment: str = "") -> None:
        now = _utc_now()
        expires_at = now + timedelta(days=days)
        with self._connect() as conn:
            conn.execute(
                """
                INSERT INTO suppressed_runs (run_id, created_at, expires_at, comment)
                VALUES (?, ?, ?, ?)
                ON CONFLICT(run_id) DO UPDATE SET
                  expires_at = excluded.expires_at,
                  comment = excluded.comment
                """,
                (run_id, now.isoformat(), expires_at.isoformat(), comment),
            )

    def filter_suppressed(self, findings: Iterable[ScoutFinding]) -> List[ScoutFinding]:
        now = _utc_now()
        filtered: List[ScoutFinding] = []

        for finding in findings:
            if not finding.evidence:
                filtered.append(finding)
                continue

            hashes = [
                _finding_hash(finding.scout_name, row.record_id, row.reason_code)
                for row in finding.evidence
                if row.record_id
            ]

            suppressed: set[str] = set()
            with self._connect() as conn:
                for chunk in _chunks(hashes):
                    placeholders = ",".join("?" for _ in chunk)
                    query = (
                        "SELECT finding_hash, expires_at FROM suppressed_findings "
                        f"WHERE finding_hash IN ({placeholders})"
                    )
                    for row in conn.execute(query, chunk):
                        expires_at = _parse_iso(row["expires_at"])
                        if expires_at and expires_at > now:
                            suppressed.add(row["finding_hash"])

            new_rows: List[EvidenceRow] = []
            for row in finding.evidence:
                if not row.record_id:
                    continue
                hash_value = _finding_hash(finding.scout_name, row.record_id, row.reason_code)
                if hash_value in suppressed:
                    continue
                new_rows.append(row)

            filtered.append(
                replace(
                    finding,
                    evidence=new_rows,
                    findings_count=len(new_rows),
                    status="findings" if new_rows else "ok",
                )
            )

        return filtered

    def cleanup(self) -> int:
        now = _utc_now().isoformat()
        with self._connect() as conn:
            cursor = conn.execute(
                "DELETE FROM suppressed_findings WHERE expires_at <= ?",
                (now,),
            )
            conn.execute(
                "DELETE FROM suppressed_runs WHERE expires_at <= ?",
                (now,),
            )
            return cursor.rowcount or 0


__all__ = ["SuppressStore"]



import unittest
from unittest.mock import patch, MagicMock
import pandas as pd
import numpy as np
from preprocessing.config.preprocess_config import PreprocessConfig

class TestB2CIntegration(unittest.TestCase):
    """Ensure new preprocessing doesn't break existing B2C"""
    
    def setUp(self):
        # Test data matching your production scenario
        self.test_df = pd.DataFrame({
            'company_name': ['Acme Inc', 'TechCorp', 'DataCo', 'Analytics LLC'],
            'email': ['john@gmail.com', 'info@techcorp.com', 'data@yahoo.com', 'sales@analytics.com'],
            'domain': ['gmail.com', 'techcorp.com', 'yahoo.com', 'analytics.com']
        })
        
    def test_b2c_function_signature_unchanged(self):
        """Ensure we're calling cleanse_domain_field correctly"""
        from fmatch.core.b2c_engine_integration import cleanse_domain_field
        
        # It should return Series, not dict
        result = cleanse_domain_field(self.test_df['domain'])
        self.assertIsInstance(result, pd.Series)
        self.assertEqual(len(result), len(self.test_df))
        
    def test_b2c_filtering_still_works(self):
        """Ensure B2C domains are still filtered with company fallback"""
        from preprocessing.core.pre_domains import DomainFilter
        from preprocessing.config.preprocess_config import PreprocessConfig
        
        config = PreprocessConfig(strip_b2c_domains=True)
        filter = DomainFilter()
        
        result_df, metrics = filter.apply_domain_filtering(self.test_df.copy(), config)
        
        # B2C domains use company fallback (normalized: lowercase, suffix removed)
        expected = ['acme', 'techcorp.com', 'dataco', 'analytics.com']
        self.assertEqual(result_df['domain'].tolist(), expected)
        
    def test_b2c_filtering_without_fallback(self):
        """Ensure B2C domains are filtered to empty when no company fallback"""
        from preprocessing.core.pre_domains import DomainFilter
        from preprocessing.config.preprocess_config import PreprocessConfig
        
        # Test data WITHOUT company column
        test_df_no_company = pd.DataFrame({
            'domain': ['gmail.com', 'techcorp.com', 'yahoo.com', 'analytics.com']
        })
        
        config = PreprocessConfig(strip_b2c_domains=True)
        filter = DomainFilter()
        
        result_df, metrics = filter.apply_domain_filtering(test_df_no_company.copy(), config)
        
        # Without company fallback, B2C domains should be empty
        self.assertEqual(result_df['domain'].tolist(), ['', 'techcorp.com', '', 'analytics.com'])
        
    def test_company_fallback_preserved(self):
        """Ensure company name fallback still works"""
        from fmatch.core.b2c_engine_integration import cleanse_domain_field
        
        result = cleanse_domain_field(
            self.test_df['email'],
            strip_b2c=True,
            extract_from_email=True,
            company_name_fallback=self.test_df['company_name']
        )
        
        # B2C implementation normalizes company names
        expected = ['acme', 'techcorp.com', 'dataco', 'analytics.com']
        self.assertEqual(result.tolist(), expected)
        
    def test_gui_config_compatibility(self):
        """Ensure GUI configuration structure still works"""
        # Simulate GUI config structure
        gui_config = {
            'enabled': True,
            'extract_from_email': True,
            'company_fallback_enabled': True,
            'stats_tracking': True
        }
        
        config = PreprocessConfig(
            strip_b2c_domains=gui_config['enabled'],
            b2c_config=gui_config
        )
        
        self.assertTrue(config.b2c_enabled)  # Using property alias
        self.assertEqual(config.b2c_config, gui_config)

if __name__ == '__main__':
    unittest.main()


"""Background worker to prefetch SFDC Accounts into DuckDB cache."""

from __future__ import annotations

import asyncio
import json
import logging
from typing import Any, Dict, Optional, Tuple
from uuid import uuid4

from ..services.sfdc_account_cache import SfdcAccountCache
from ..services.salesforce_gateway import (
    SalesforceAuthError,
    SalesforceConnectionError,
    SalesforceGateway,
    SalesforceRateLimitError,
)

logger = logging.getLogger(__name__)

STATE_KEY = "sfdc_prefetch:job:{job_id}:state"
PROGRESS_KEY = "sfdc_prefetch:job:{job_id}:progress"
RESULT_KEY = "sfdc_prefetch:job:{job_id}:result"
IDEMPOTENCY_KEY = "sfdc_prefetch:{account_id}:{key}"

RESULT_TTL = 86400  # 24h
LOCK_WAIT_S = 300


class SfdcAccountPrefetchWorker:
    """Async worker that warms the DuckDB account cache via Bulk API."""

    def __init__(self, redis=None) -> None:
        self.redis = redis

    async def start_prefetch_job(
        self,
        account_id: str,
        integration_id: str,
        gateway: SalesforceGateway,
        idempotency_key: Optional[str] = None,
    ) -> Tuple[str, bool]:
        if self.redis and idempotency_key:
            cache_key = IDEMPOTENCY_KEY.format(
                account_id=account_id, key=idempotency_key
            )
            existing = await self.redis.get(cache_key)
            if existing:
                job_id_existing = (
                    existing.decode("utf-8")
                    if hasattr(existing, "decode")
                    else str(existing)
                )
                return job_id_existing, False

        job_id = str(uuid4())
        if self.redis:
            await self.redis.set(STATE_KEY.format(job_id=job_id), "queued", ex=RESULT_TTL)
            await self._set_progress(job_id, 0, "queued")
            if idempotency_key:
                await self.redis.set(
                    IDEMPOTENCY_KEY.format(account_id=account_id, key=idempotency_key),
                    job_id,
                    ex=RESULT_TTL,
                )

        asyncio.create_task(
            self._run_job(
                job_id=job_id,
                account_id=account_id,
                integration_id=integration_id,
                gateway=gateway,
            )
        )
        return job_id, True

    async def _run_job(
        self,
        job_id: str,
        account_id: str,
        integration_id: str,
        gateway: SalesforceGateway,
    ) -> None:
        await self._set_state(job_id, "processing")
        cache = SfdcAccountCache(
            account_id=account_id,
            integration_id=integration_id,
            redis=self.redis,
        )

        async def _progress(pct: int, phase: str) -> None:
            await self._set_progress(job_id, pct, phase)

        try:
            await self._set_progress(job_id, 5, "prefetching")
            await cache.get_or_prefetch(
                gateway, force_refresh=True, progress_callback=_progress
            )
            # get_or_prefetch returns bool; read meta for count
            meta = await cache.get_metadata()
            account_count = meta.get("account_count") if meta else None
            await self._set_result(
                job_id,
                {
                    "accounts_fetched": account_count,
                    "cache_path": str(cache.db_path),
                },
            )
            await self._set_state(job_id, "completed")
        except SalesforceRateLimitError as exc:
            await self._fail(job_id, f"RATE_LIMITED: {exc}")
        except SalesforceAuthError as exc:
            await self._fail(job_id, f"SFDC_AUTH_REQUIRED: {exc}")
        except SalesforceConnectionError as exc:
            await self._fail(job_id, f"CONNECTION_ERROR: {exc}")
        except Exception as exc:  # noqa: BLE001
            logger.exception("Prefetch job %s failed", job_id)
            await self._fail(job_id, str(exc))

    # ─────────────── Redis helpers ───────────────
    async def _set_state(self, job_id: str, state: str) -> None:
        if not self.redis:
            return
        await self.redis.set(STATE_KEY.format(job_id=job_id), state, ex=RESULT_TTL)

    async def _set_progress(self, job_id: str, pct: int, phase: str) -> None:
        if not self.redis:
            return
        payload = json.dumps({"percent": pct, "phase": phase})
        await self.redis.set(PROGRESS_KEY.format(job_id=job_id), payload, ex=RESULT_TTL)

    async def _set_result(self, job_id: str, payload: Dict[str, Any]) -> None:
        if not self.redis:
            return
        await self.redis.set(
            RESULT_KEY.format(job_id=job_id),
            json.dumps(payload),
            ex=RESULT_TTL,
        )

    async def _fail(self, job_id: str, message: str) -> None:
        await self._set_state(job_id, "failed")
        await self._set_result(job_id, {"error": message})
        await self._set_progress(job_id, 100, "failed")



from fmatch.saas.operators.fix_phone_format import fix_phone_format


def test_us_phone_basic():
    assert fix_phone_format("(415) 555-1212", country="US")["value"] == "+14155551212"
    assert fix_phone_format("(415) 555-1212 x123", country="US")["value"] == "+14155551212"
    assert fix_phone_format("123", country="US")["value"] == ""




"""
Salesforce pack orchestration for data changes (plan/apply/drift).
"""

from __future__ import annotations

import hashlib
import json
from dataclasses import dataclass
from datetime import date, datetime, timezone
from pathlib import Path
from typing import Any, Dict, List, Optional

from g_gremlin import sfdc as sfdc_helpers
from g_gremlin import sfdc_schema as sfdc_schema_helpers
from g_gremlin.errors import missing_pyyaml_error
from g_gremlin.pacing import BatchPacer, get_profile, PacingProfile


@dataclass
class PackPlan:
    payload: Dict[str, Any]

    def to_dict(self) -> Dict[str, Any]:
        return self.payload


@dataclass
class ApplyResult:
    payload: Dict[str, Any]

    def to_dict(self) -> Dict[str, Any]:
        return self.payload


@dataclass
class DriftResult:
    payload: Dict[str, Any]

    def to_dict(self) -> Dict[str, Any]:
        return self.payload


def init_pack(dir_path: Path) -> None:
    dir_path.mkdir(parents=True, exist_ok=True)
    (dir_path / "artifacts").mkdir(exist_ok=True)
    (dir_path / "data").mkdir(exist_ok=True)

    pack_yaml = dir_path / "pack.yaml"
    if not pack_yaml.exists():
        pack_yaml.write_text(
            "\n".join(
                [
                    "name: sfdc_data_pack",
                    "object: Account",
                    "csv: ./data/updates.csv",
                    "id_column: Id",
                    "key_field: Id",
                    "fields:",
                    "  - Name",
                    "  - Industry",
                    "",
                    "options:",
                    "  mode: safe",
                    "  batch_size: 50",
                    "  inter_batch_delay: 1.0",
                    "  allow_blank_fields: false",
                    "  deny_nulls: false",
                    "  allowlist_fields: []",
                    "  max_updates: null",
                    "  max_plan_age_minutes: null",
                    "  require_key_field: true",
                    "  snapshot_dir: null",
                    "  smoke_after_apply: false",
                    "",
                    "smoke: []",
                    "",
                ]
            ),
            encoding="utf-8",
        )

    readme = dir_path / "README.md"
    if not readme.exists():
        readme.write_text(
            "\n".join(
                [
                    "# Salesforce Pack",
                    "",
                    "Workflow:",
                    "1) g-gremlin sfdc snapshot --object Account --fields Id,Name --out snapshots/baseline",
                    "2) g-gremlin sfdc pack plan .",
                    "3) g-gremlin sfdc pack apply . --yes",
                    "4) g-gremlin sfdc drift .",
                    "5) g-gremlin sfdc smoke .",
                    "",
                ]
            ),
            encoding="utf-8",
        )


def load_pack_yaml(path: Path) -> Dict[str, Any]:
    if not path.exists():
        raise ValueError(f"Pack file not found: {path}")
    raw_text = path.read_text(encoding="utf-8")
    if path.suffix.lower() in {".yaml", ".yml"}:
        try:
            import yaml  # type: ignore
        except Exception as exc:
            raise missing_pyyaml_error() from exc
        return yaml.safe_load(raw_text) or {}
    return json.loads(raw_text)


def plan_pack(pack_dir: Path, *, org_alias: Optional[str] = None) -> PackPlan:
    pack = load_pack_yaml(pack_dir / "pack.yaml")
    object_name = str(pack.get("object") or pack.get("sobject") or "").strip()
    if not object_name:
        raise ValueError("pack.yaml missing object")
    sfdc_helpers.ensure_valid_identifier(object_name, "object name")
    csv_path = _resolve_path(pack_dir, pack.get("csv"))
    id_column = str(pack.get("id_column") or "Id")
    key_field = str(pack.get("key_field") or "Id").strip()
    fields = _parse_fields(pack.get("fields") or [])
    options = dict(pack.get("options") or {})
    if options.get("snapshot_dir"):
        resolved = _resolve_path(pack_dir, options.get("snapshot_dir"))
        if resolved:
            options["snapshot_dir"] = str(resolved)

    if not csv_path or not csv_path.exists():
        raise ValueError("CSV path not found. Set pack.yaml csv to an existing file.")
    if not fields:
        raise ValueError("pack.yaml requires at least one field.")
    if not key_field:
        raise ValueError("pack.yaml missing key_field (default is Id).")
    sfdc_helpers.ensure_valid_identifier(key_field, "key_field")
    require_key_field = bool(options.get("require_key_field", True))
    if require_key_field and id_column.lower() != "id" and key_field.lower() == "id":
        raise ValueError("id_column is not Id; set key_field to the Salesforce external ID field.")
    if require_key_field and key_field.lower() != "id" and id_column.lower() == "id":
        raise ValueError("key_field is external; set id_column to the matching CSV column.")
    if require_key_field and id_column.lower() != key_field.lower():
        if id_column.lower() != "id" and key_field.lower() != "id":
            raise ValueError("id_column must match key_field when using an external ID key.")

    rows, headers = sfdc_helpers.read_csv(csv_path)
    if id_column not in headers:
        raise ValueError(f"ID column '{id_column}' not found in {csv_path}.")

    field_columns = {f["field"]: f["column"] for f in fields}
    for f in fields:
        if f["column"] not in headers:
            raise ValueError(f"Column '{f['column']}' not found in {csv_path}.")

    allow_blank = bool(options.get("allow_blank_fields", False))
    deny_nulls = bool(options.get("deny_nulls", False))
    allowlist_fields = options.get("allowlist_fields") or []
    if isinstance(allowlist_fields, str):
        allowlist_fields = [f.strip() for f in allowlist_fields.split(",") if f.strip()]
    max_updates = options.get("max_updates")

    id_map, duplicates = _build_id_map(rows, id_column, key_field)
    if duplicates:
        raise ValueError(f"Duplicate keys found in '{id_column}': {len(duplicates)} duplicates.")
    records = _fetch_records(
        object_name,
        list(id_map.values()),
        fields,
        key_field=key_field,
        org_alias=org_alias,
    )

    changes, not_found = _build_changes(
        rows,
        id_column,
        fields,
        field_columns,
        records,
        key_field=key_field,
        allow_blank=allow_blank,
    )

    summary = {
        "total_records": len(id_map),
        "records_with_changes": len(changes),
        "records_unchanged": max(len(id_map) - len(changes) - len(not_found), 0),
        "records_not_found": len(not_found),
    }

    validation_errors: List[str] = []
    validation_warnings: List[str] = []

    if allowlist_fields:
        allowlist_fields = [str(f).strip() for f in allowlist_fields if str(f).strip()]
        for field in allowlist_fields:
            sfdc_helpers.ensure_valid_identifier(field, "allowlist field")
        for field in [f["field"] for f in fields]:
            if field not in allowlist_fields:
                validation_errors.append(f"Field '{field}' not in allowlist_fields.")

    if max_updates is not None:
        if len(changes) > int(max_updates):
            validation_errors.append(
                f"max_updates exceeded: {len(changes)} updates > {int(max_updates)} limit."
            )

    toolchain = _toolchain_fingerprint()
    auth = _auth_fingerprint(org_alias)
    input_hash = _sha256_file(csv_path)
    baseline_hash = _sha256_json({"changes": changes, "not_found": not_found})

    identity = sfdc_helpers.get_org_identity(org_alias=org_alias)
    schema_info = sfdc_schema_helpers.get_schema(
        object_name,
        identity=identity,
        org_alias=org_alias,
    )
    schema_warnings = schema_info.warnings or []
    validation_warnings.extend(schema_warnings)
    if key_field and key_field not in schema_info.fields:
        validation_errors.append(f"key_field '{key_field}' not found in schema.")
    for field in [f["field"] for f in fields]:
        schema_field = schema_info.fields.get(field)
        if not schema_field:
            validation_errors.append(f"Field '{field}' not found in schema.")
            continue
        if not schema_field.updateable:
            validation_errors.append(f"Field '{field}' is not updateable.")
    schema_errors, schema_warns, changes = _validate_changes(
        changes,
        schema_info.fields,
        deny_nulls=deny_nulls,
    )
    validation_errors.extend(schema_errors)
    validation_warnings.extend(schema_warns)
    validation_errors = _dedupe_messages(validation_errors)
    validation_warnings = _dedupe_messages(validation_warnings)

    plan_payload = {
        "$schema": "SfdcPackPlan/v1",
        "plan_id": _new_plan_id(),
        "created_at": datetime.now(timezone.utc).isoformat(),
        "pack_name": pack.get("name") or pack_dir.name,
        "object": object_name,
        "key_field": key_field,
        "csv": str(csv_path),
        "id_column": id_column,
        "fields": fields,
        "summary": summary,
        "changes": changes,
        "not_found": not_found,
        "baseline_fingerprint": baseline_hash,
        "input_fingerprint": input_hash,
        "toolchain_fingerprint": toolchain,
        "auth_fingerprint": auth,
        "schema": {
            "schema_hash": schema_info.schema_hash,
            "capabilities_fingerprint": schema_info.metadata.get("capabilities_fingerprint"),
            "cache_dir": str(schema_info.cache_dir),
        },
        "validation": {
            "blocked": bool(validation_errors),
            "errors": validation_errors,
            "warnings": validation_warnings,
        },
        "options": options,
    }
    plan_payload["plan_hash"] = _compute_data_plan_hash(plan_payload)

    artifact_dir = _artifact_dir(pack_dir)
    artifact_dir.mkdir(parents=True, exist_ok=True)
    plan_path = artifact_dir / "plan.json"
    plan_path.write_text(json.dumps(plan_payload, indent=2), encoding="utf-8")

    return PackPlan(payload=plan_payload)


def apply_pack(
    pack_dir: Path,
    *,
    org_alias: Optional[str] = None,
    plan_path: Optional[Path] = None,
    yes: bool = False,
    replan: bool = False,
    expected_plan_hash: str | None = None,
) -> ApplyResult:
    if not yes:
        raise ValueError("Refusing to apply without --yes.")

    pack = load_pack_yaml(pack_dir / "pack.yaml")
    object_name = str(pack.get("object") or pack.get("sobject") or "").strip()
    if not object_name:
        raise ValueError("pack.yaml missing object")
    sfdc_helpers.ensure_valid_identifier(object_name, "object name")
    csv_path = _resolve_path(pack_dir, pack.get("csv"))
    id_column = str(pack.get("id_column") or "Id")
    fields = _parse_fields(pack.get("fields") or [])
    options = pack.get("options") or {}

    if not plan_path:
        plan_path = _latest_plan(pack_dir)
    if not plan_path or not plan_path.exists():
        raise ValueError("No plan found. Run: g-gremlin sfdc pack plan <dir>")

    plan = json.loads(plan_path.read_text(encoding="utf-8"))
    stored_hash = str(plan.get("plan_hash") or "").strip()
    if stored_hash:
        recomputed = _compute_data_plan_hash(plan)
        if recomputed != stored_hash:
            raise ValueError("Plan hash mismatch -- plan file may have been tampered with.")
    if expected_plan_hash:
        expected = str(expected_plan_hash).strip()
        if not stored_hash:
            raise ValueError("Plan hash missing from plan file; cannot verify --plan-hash.")
        if stored_hash != expected:
            raise ValueError("Provided --plan-hash does not match plan file hash.")

    validation = plan.get("validation") or {}
    if validation.get("blocked"):
        raise ValueError("Plan is blocked due to validation errors. Re-run plan after fixing errors.")
    plan_opts = plan.get("options") or {}
    max_plan_age = plan_opts.get("max_plan_age_minutes")
    if max_plan_age:
        created_at = plan.get("created_at")
        if created_at:
            created_dt = datetime.fromisoformat(created_at)
            age_minutes = (datetime.now(timezone.utc) - created_dt).total_seconds() / 60.0
            if age_minutes > float(max_plan_age):
                raise ValueError(
                    f"Plan is {age_minutes:.1f} minutes old; exceeds max_plan_age_minutes={max_plan_age}."
                )
    drift = detect_drift(pack_dir, plan, org_alias=org_alias)
    if drift["has_drift"]:
        if not replan:
            raise ValueError("Plan drift detected. Re-run plan or pass --replan.")
        replanned = plan_pack(pack_dir, org_alias=org_alias)
        plan = replanned.payload
        drift = detect_drift(pack_dir, plan, org_alias=org_alias)
        if drift["has_drift"]:
            raise ValueError("Drift still detected after replan.")

    if not csv_path or not csv_path.exists():
        raise ValueError("CSV path not found. Set pack.yaml csv to an existing file.")

    changes = plan.get("changes") or []
    if not changes:
        return ApplyResult(
            payload={
                "$schema": "SfdcPackApply/v1",
                "status": "noop",
                "summary": {"updated": 0, "failed": 0},
                "plan_id": plan.get("plan_id"),
                "artifact_dir": None,
            }
        )

    pacing = _resolve_pacing(plan_opts or options)
    pacer = BatchPacer(pacing)

    successes: List[Dict[str, Any]] = []
    failures: List[Dict[str, Any]] = []

    artifact_dir = _artifact_dir(pack_dir)
    artifact_dir.mkdir(parents=True, exist_ok=True)
    receipts_dir = artifact_dir / "receipts"
    receipts_dir.mkdir(exist_ok=True)
    (artifact_dir / "plan_used.json").write_text(json.dumps(plan, indent=2), encoding="utf-8")
    (artifact_dir / "drift_check.json").write_text(json.dumps(drift, indent=2), encoding="utf-8")

    before_rows = _receipt_rows_from_plan(changes)
    _write_receipt_csv(receipts_dir / "before.csv", before_rows)

    for batch in pacer.chunk(changes):
        for change in batch:
            record_id = change.get("id") or ""
            fields_payload = change.get("fields") or {}
            update_values = {field: meta.get("new") for field, meta in fields_payload.items()}
            try:
                result = sfdc_helpers.update_record(
                    object_name,
                    record_id,
                    update_values,
                    org_alias=org_alias,
                )
                successes.append({"id": record_id, "result": result})
            except Exception as exc:
                failures.append({"id": record_id, "error": str(exc)})
        pacer.pace()

    after_rows = _fetch_after_rows(object_name, changes, org_alias=org_alias)
    _write_receipt_csv(receipts_dir / "after.csv", after_rows)
    _write_jsonl(receipts_dir / "success.jsonl", successes)
    _write_jsonl(receipts_dir / "errors.jsonl", failures)

    summary = {
        "updated": len(successes),
        "failed": len(failures),
        "total": len(changes),
    }
    payload = {
        "$schema": "SfdcPackApply/v1",
        "status": "success" if not failures else "partial",
        "summary": summary,
        "plan_id": plan.get("plan_id"),
        "plan_path": str(plan_path) if plan_path else None,
        "artifact_dir": str(artifact_dir),
    }
    (artifact_dir / "apply.json").write_text(json.dumps(payload, indent=2), encoding="utf-8")
    return ApplyResult(payload=payload)


def detect_drift(
    pack_dir: Path,
    plan: Dict[str, Any],
    *,
    org_alias: Optional[str] = None,
) -> Dict[str, Any]:
    changes = plan.get("changes") or []
    object_name = plan.get("object") or ""
    ids = [c.get("id") for c in changes if c.get("id")]
    drifted: List[Dict[str, Any]] = []
    snapshot_drifted: List[Dict[str, Any]] = []

    fields = plan.get("fields") or []
    current = _fetch_records(object_name, ids, fields, key_field="Id", org_alias=org_alias)
    for change in changes:
        record_id = change.get("id") or ""
        current_record = current.get(_normalize_key(record_id, "Id"))
        if not current_record:
            drifted.append({"id": record_id, "reason": "record_missing"})
            continue
        for field, meta in (change.get("fields") or {}).items():
            baseline = meta.get("current")
            live = current_record.get(field)
            if str(baseline or "") != str(live or ""):
                drifted.append(
                    {
                        "id": record_id,
                        "field": field,
                        "baseline": baseline,
                        "current": live,
                    }
                )

    options = plan.get("options") or {}
    snapshot_dir = options.get("snapshot_dir")
    if snapshot_dir:
        snapshot_baseline = _load_snapshot_baseline(Path(snapshot_dir), object_name)
        if snapshot_baseline:
            for change in changes:
                record_id = change.get("id") or ""
                baseline_row = snapshot_baseline.get(_normalize_key(record_id, "Id"))
                current_record = current.get(_normalize_key(record_id, "Id"))
                if not baseline_row or not current_record:
                    continue
                for field in (change.get("fields") or {}).keys():
                    snap_val = baseline_row.get(field)
                    live_val = current_record.get(field)
                    if str(snap_val or "") != str(live_val or ""):
                        snapshot_drifted.append(
                            {
                                "id": record_id,
                                "field": field,
                                "snapshot": snap_val,
                                "current": live_val,
                            }
                        )

    input_fingerprint = plan.get("input_fingerprint")
    csv_path = plan.get("csv")
    csv_changed = False
    if csv_path and input_fingerprint:
        try:
            csv_changed = _sha256_file(Path(csv_path)) != input_fingerprint
        except Exception:
            csv_changed = True

    return {
        "has_drift": bool(drifted) or csv_changed or bool(snapshot_drifted),
        "records_changed": drifted,
        "csv_changed": csv_changed,
        "snapshot_changed": snapshot_drifted,
        "total_records_checked": len(ids),
    }


def compute_drift(pack_dir: Path, *, org_alias: Optional[str] = None) -> DriftResult:
    plan_path = _latest_plan(pack_dir)
    if not plan_path or not plan_path.exists():
        raise ValueError("No plan found. Run: g-gremlin sfdc pack plan <dir>")
    plan = json.loads(plan_path.read_text(encoding="utf-8"))
    drift = detect_drift(pack_dir, plan, org_alias=org_alias)
    payload = {
        "$schema": "SfdcPackDrift/v1",
        "plan_id": plan.get("plan_id"),
        "has_drift": drift["has_drift"],
        "csv_changed": drift["csv_changed"],
        "records_changed": drift["records_changed"],
        "snapshot_changed": drift.get("snapshot_changed") or [],
        "total_records_checked": drift["total_records_checked"],
    }
    return DriftResult(payload=payload)


def _artifact_dir(pack_dir: Path) -> Path:
    timestamp = datetime.now(timezone.utc).strftime("%Y%m%d_%H%M%S")
    return pack_dir / "artifacts" / timestamp


def _latest_plan(pack_dir: Path) -> Optional[Path]:
    artifacts = pack_dir / "artifacts"
    if not artifacts.exists():
        return None
    dirs = sorted([d for d in artifacts.iterdir() if d.is_dir()], reverse=True)
    for d in dirs:
        plan_path = d / "plan.json"
        if plan_path.exists():
            return plan_path
    return None


def _resolve_path(pack_dir: Path, raw_path: Any) -> Optional[Path]:
    if not raw_path:
        return None
    path = Path(str(raw_path))
    if path.is_absolute():
        return path
    return (pack_dir / path).resolve()


def _parse_fields(items: List[Any]) -> List[Dict[str, str]]:
    parsed: List[Dict[str, str]] = []
    for item in items:
        if isinstance(item, str):
            field = item.strip()
            if field:
                sfdc_helpers.ensure_valid_identifier(field, "field")
                parsed.append({"field": field, "column": field})
            continue
        if isinstance(item, dict):
            field = str(item.get("field") or item.get("name") or "").strip()
            column = str(item.get("column") or item.get("csv") or field).strip()
            if field:
                sfdc_helpers.ensure_valid_identifier(field, "field")
                parsed.append({"field": field, "column": column or field})
    return parsed


def _build_id_map(
    rows: List[Dict[str, str]],
    id_column: str,
    key_field: str,
) -> tuple[Dict[str, str], List[str]]:
    mapping: Dict[str, str] = {}
    duplicates: List[str] = []
    for row in rows:
        raw_id = (row.get(id_column) or "").strip()
        if not raw_id:
            continue
        key = _normalize_key(raw_id, key_field)
        if not key:
            continue
        if key in mapping:
            duplicates.append(raw_id)
            continue
        mapping[key] = raw_id
    return mapping, duplicates


def _fetch_records(
    sobject: str,
    ids: List[str],
    fields: List[Dict[str, str]],
    *,
    key_field: str,
    org_alias: Optional[str] = None,
) -> Dict[str, Dict[str, str]]:
    records: Dict[str, Dict[str, str]] = {}
    if not ids:
        return records
    field_list = [f["field"] for f in fields]
    select_fields = ["Id"]
    if "Name" not in field_list:
        select_fields.append("Name")
    if key_field not in select_fields:
        select_fields.append(key_field)
    select_fields += field_list

    chunk_size = 150
    for i in range(0, len(ids), chunk_size):
        chunk = ids[i : i + chunk_size]
        soql_ids = ", ".join(f"'{sfdc_helpers.escape_soql_value(val)}'" for val in chunk)
        soql = f"SELECT {', '.join(select_fields)} FROM {sobject} WHERE {key_field} IN ({soql_ids})"
        rows = sfdc_helpers.run_soql(soql, org_alias=org_alias)
        for record in rows:
            key_val = record.get(key_field, "") if key_field else record.get("Id", "")
            key = _normalize_key(str(key_val), key_field)
            if key:
                records[key] = record
    return records


def _build_changes(
    rows: List[Dict[str, str]],
    id_column: str,
    fields: List[Dict[str, str]],
    field_columns: Dict[str, str],
    records: Dict[str, Dict[str, str]],
    *,
    key_field: str,
    allow_blank: bool,
) -> tuple[List[Dict[str, Any]], List[str]]:
    changes: List[Dict[str, Any]] = []
    not_found: List[str] = []

    for row in rows:
        raw_id = (row.get(id_column) or "").strip()
        if not raw_id:
            continue
        key = _normalize_key(raw_id, key_field)
        current = records.get(key)
        if not current:
            if raw_id not in not_found:
                not_found.append(raw_id)
            continue
        field_changes: Dict[str, Dict[str, Any]] = {}
        for field in fields:
            field_name = field["field"]
            column = field_columns[field_name]
            new_val = _clean_val(row.get(column))
            current_val = _clean_val(current.get(field_name))
            if new_val in (None, "") and not allow_blank:
                continue
            if _values_different(current_val, new_val):
                field_changes[field_name] = {"current": current_val, "new": new_val}
        if field_changes:
            changes.append(
                {
                    "id": current.get("Id", raw_id),
                    "name": current.get("Name"),
                    "key_value": raw_id,
                    "fields": field_changes,
                }
            )
    return changes, not_found


def _resolve_pacing(options: Dict[str, Any]) -> PacingProfile:
    mode = str(options.get("mode") or "safe").lower().strip()
    base = get_profile(mode)
    profile = PacingProfile(
        name=base.name,
        batch_size=base.batch_size,
        inter_batch_delay=base.inter_batch_delay,
        rate_limit_headroom=base.rate_limit_headroom,
    )
    batch_size = options.get("batch_size")
    inter_delay = options.get("inter_batch_delay")
    if batch_size:
        profile.batch_size = int(batch_size)
    if inter_delay is not None:
        profile.inter_batch_delay = float(inter_delay)
    return profile


def _receipt_rows_from_plan(changes: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    rows: List[Dict[str, Any]] = []
    for change in changes:
        row = {"Id": change.get("id"), "Name": change.get("name")}
        if change.get("key_value") is not None:
            row["KeyValue"] = change.get("key_value")
        for field, meta in (change.get("fields") or {}).items():
            row[field] = meta.get("current")
        rows.append(row)
    return rows


def _fetch_after_rows(
    sobject: str,
    changes: List[Dict[str, Any]],
    *,
    org_alias: Optional[str],
) -> List[Dict[str, Any]]:
    ids = [c.get("id") for c in changes if c.get("id")]
    change_fields = _fields_from_changes(changes)
    fields = [{"field": f, "column": f} for f in change_fields]
    records = _fetch_records(sobject, ids, fields, key_field="Id", org_alias=org_alias)
    rows: List[Dict[str, Any]] = []
    for change in changes:
        record_id = change.get("id")
        record = records.get(_normalize_key(record_id, "Id"))
        if not record:
            continue
        row = {"Id": record.get("Id"), "Name": record.get("Name")}
        if change.get("key_value") is not None:
            row["KeyValue"] = change.get("key_value")
        for field in change_fields:
            row[field] = record.get(field)
        rows.append(row)
    return rows


def _fields_from_changes(changes: List[Dict[str, Any]]) -> List[str]:
    return list(
        dict.fromkeys(
            field
            for change in changes
            for field in (change.get("fields") or {}).keys()
        )
    )


def _write_receipt_csv(path: Path, rows: List[Dict[str, Any]]) -> None:
    if not rows:
        return
    headers = list(rows[0].keys())
    with path.open("w", newline="", encoding="utf-8") as handle:
        import csv as csv_module

        writer = csv_module.DictWriter(handle, fieldnames=headers)
        writer.writeheader()
        for row in rows:
            writer.writerow({h: row.get(h, "") for h in headers})


def _write_jsonl(path: Path, rows: List[Dict[str, Any]]) -> None:
    if not rows:
        return
    with path.open("w", encoding="utf-8") as handle:
        for row in rows:
            handle.write(json.dumps(row) + "\n")


def _toolchain_fingerprint() -> Dict[str, Any]:
    fingerprint = {}
    try:
        fingerprint["sf_version"] = sfdc_helpers.get_cli_version()
    except Exception as exc:
        fingerprint["sf_version_error"] = str(exc)
    try:
        fingerprint["data_query_available"] = sfdc_helpers.check_command_available(["data", "query"])
        fingerprint["data_update_available"] = sfdc_helpers.check_command_available(["data", "update", "record"])
        describe_cmd = sfdc_helpers.get_describe_command_path()
        fingerprint["schema_describe_available"] = describe_cmd is not None
        if describe_cmd:
            fingerprint["schema_describe_command"] = " ".join(describe_cmd)
    except Exception as exc:
        fingerprint["command_check_error"] = str(exc)
    return fingerprint


def _auth_fingerprint(org_alias: Optional[str]) -> Dict[str, Any]:
    try:
        return sfdc_helpers.get_org_identity(org_alias=org_alias)
    except Exception as exc:
        return {"error": str(exc)}


def _sha256_file(path: Path) -> str:
    digest = hashlib.sha256()
    with path.open("rb") as handle:
        for chunk in iter(lambda: handle.read(8192), b""):
            digest.update(chunk)
    return digest.hexdigest()


def _sha256_json(payload: Dict[str, Any]) -> str:
    raw = json.dumps(payload, sort_keys=True).encode("utf-8")
    return hashlib.sha256(raw).hexdigest()


def _compute_data_plan_hash(plan_payload: Dict[str, Any]) -> str:
    canonical = {
        "plan_id": plan_payload.get("plan_id"),
        "object": plan_payload.get("object"),
        "key_field": plan_payload.get("key_field"),
        "fields": plan_payload.get("fields") or [],
        "changes": plan_payload.get("changes") or [],
        "not_found": plan_payload.get("not_found") or [],
        "baseline_fingerprint": plan_payload.get("baseline_fingerprint"),
        "input_fingerprint": plan_payload.get("input_fingerprint"),
    }
    raw = json.dumps(canonical, sort_keys=True, separators=(",", ":")).encode("utf-8")
    return hashlib.sha256(raw).hexdigest()


def _new_plan_id() -> str:
    return datetime.now(timezone.utc).strftime("plan-%Y%m%d-%H%M%S")


def _dedupe_messages(items: List[str]) -> List[str]:
    seen: set[str] = set()
    unique: List[str] = []
    for item in items:
        if item in seen:
            continue
        seen.add(item)
        unique.append(item)
    return unique

def _normalize_key(value: Any, key_field: str) -> str:
    if value is None:
        return ""
    text = str(value).strip()
    if key_field.lower() == "id":
        return sfdc_helpers.match_key(text)
    return text

def _clean_val(value: Any) -> Optional[str]:
    if value is None:
        return None
    text = str(value).strip()
    return text if text != "" else None


def _values_different(current: Optional[str], new: Optional[str]) -> bool:
    return (current or "") != (new or "")


def _validate_changes(
    changes: List[Dict[str, Any]],
    field_map: Dict[str, sfdc_schema_helpers.FieldInfo],
    *,
    deny_nulls: bool,
) -> tuple[List[str], List[str], List[Dict[str, Any]]]:
    errors: List[str] = []
    warnings: List[str] = []
    updated_changes: List[Dict[str, Any]] = []

    for change in changes:
        updated_fields: Dict[str, Dict[str, Any]] = {}
        for field, meta in (change.get("fields") or {}).items():
            schema = field_map.get(field)
            if not schema:
                errors.append(f"Field '{field}' not found in schema.")
                continue
            if not schema.updateable:
                errors.append(f"Field '{field}' is not updateable.")
            new_val = meta.get("new")
            if new_val in (None, ""):
                if deny_nulls or not schema.nillable:
                    errors.append(f"Field '{field}' cannot be set to null/blank.")
                updated_fields[field] = meta
                continue

            coerced, warn, err = _coerce_value(new_val, schema)
            if err:
                errors.append(f"{field}: {err}")
                continue
            if warn:
                warnings.append(f"{field}: {warn}")
            updated_fields[field] = {"current": meta.get("current"), "new": coerced}

        if updated_fields:
            updated_change = dict(change)
            updated_change["fields"] = updated_fields
            updated_changes.append(updated_change)
        else:
            updated_changes.append(change)

    return errors, warnings, updated_changes


def _coerce_value(value: Any, schema: sfdc_schema_helpers.FieldInfo) -> tuple[Any, str, str]:
    text = str(value).strip()
    ftype = (schema.type or "").lower()

    if ftype in {"boolean"}:
        truthy = {"true", "1", "yes", "y", "t"}
        falsy = {"false", "0", "no", "n", "f"}
        lowered = text.lower()
        if lowered in truthy:
            return True, "", ""
        if lowered in falsy:
            return False, "", ""
        return value, "", f"invalid boolean value '{text}'"

    if ftype in {"int"}:
        try:
            return int(text), "", ""
        except ValueError:
            return value, "", f"invalid int value '{text}'"

    if ftype in {"double", "currency", "percent"}:
        try:
            float_val = float(text)
            return float_val, "", ""
        except ValueError:
            return value, "", f"invalid number value '{text}'"

    if ftype == "date":
        try:
            parsed = date.fromisoformat(text)
            return parsed.isoformat(), "", ""
        except ValueError:
            return value, "", f"invalid date '{text}' (expected YYYY-MM-DD)"

    if ftype == "datetime":
        try:
            iso_text = text[:-1] + "+00:00" if text.endswith("Z") else text
            parsed = datetime.fromisoformat(iso_text)
            warn = ""
            if parsed.tzinfo is None:
                warn = "datetime has no timezone; using local/naive format"
            return parsed.isoformat(), warn, ""
        except ValueError:
            return value, "", f"invalid datetime '{text}'"

    if ftype == "picklist":
        if schema.restricted_picklist and text not in schema.picklist_values:
            return value, "", f"value '{text}' not in restricted picklist"
        if schema.picklist_values and text not in schema.picklist_values:
            return value, f"value '{text}' not in picklist; will attempt anyway", ""
        return value, "", ""

    if ftype == "multipicklist":
        parts = [p.strip() for p in text.split(";") if p.strip()]
        if schema.restricted_picklist:
            for part in parts:
                if part not in schema.picklist_values:
                    return value, "", f"multipicklist value '{part}' not allowed"
        return value, "", ""

    # Default: no coercion
    return value, "", ""


def _load_snapshot_baseline(snapshot_dir: Path, object_name: str) -> Dict[str, Dict[str, Any]]:
    data_path = snapshot_dir / "data" / f"{object_name}.json"
    if not data_path.exists():
        return {}
    try:
        rows = json.loads(data_path.read_text(encoding="utf-8"))
    except Exception:
        return {}
    baseline: Dict[str, Dict[str, Any]] = {}
    if isinstance(rows, list):
        for row in rows:
            if not isinstance(row, dict):
                continue
            record_id = row.get("Id") or row.get("id")
            if record_id:
                baseline[_normalize_key(str(record_id), "Id")] = row
    return baseline



from __future__ import annotations

import os
import boto3
from botocore.config import Config
from fastapi import UploadFile
from typing import Optional


class S3Backend:
    """
    Minimal async-friendly wrapper around boto3 for object storage.

    Environment:
    - STORAGE_PROVIDER=local|s3|minio (default local)
    - AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY, AWS_REGION
    - S3_BUCKET
    - S3_ENDPOINT_URL (for MinIO/custom endpoints)
    - SIGNED_URL_TTL_SEC (default 86400)
    """

    def __init__(
        self,
        *,
        bucket: Optional[str] = None,
        endpoint_url: Optional[str] = None,
        region: Optional[str] = None,
    ):
        self.bucket = bucket or os.getenv("S3_BUCKET") or os.getenv("S3_BUCKET_NAME")
        self.endpoint_url = endpoint_url or os.getenv("S3_ENDPOINT_URL")
        self.region = region or os.getenv("AWS_REGION", "us-east-1")

        # Configure boto3 client; credentials are auto-resolved from env/instance
        self._client = boto3.client(
            "s3",
            region_name=self.region,
            endpoint_url=self.endpoint_url,
            config=Config(s3={"addressing_style": "path"})
            if self.endpoint_url
            else None,
        )

    async def save(self, file: UploadFile, key: str) -> str:
        """Save object to bucket and return s3:// URL."""
        # boto3 is sync; call directly (OK for short uploads). For larger files, consider threadpool.
        body = await file.read()
        self._client.put_object(
            Bucket=self.bucket, Key=key, Body=body, ContentType=file.content_type
        )
        return f"s3://{self.bucket}/{key}"

    async def read(self, key: str) -> bytes:
        obj = self._client.get_object(Bucket=self.bucket, Key=key)
        return obj["Body"].read()

    async def generate_presigned_url(
        self, key: str, expires_in: Optional[int] = None, operation: str = "get_object"
    ) -> str:
        ttl = int(expires_in or os.getenv("SIGNED_URL_TTL_SEC", 86400))
        return self._client.generate_presigned_url(
            ClientMethod=operation,
            Params={"Bucket": self.bucket, "Key": key},
            ExpiresIn=ttl,
        )



"""Slack notification helpers for Autopilot."""

from __future__ import annotations

import os
from typing import Any, Dict

from .contracts import PlanV1, ReceiptV1
from .slack_blocks import build_plan_approval_blocks

try:  # Optional dependency in some environments.
    from slack_sdk import WebClient
except Exception as exc:  # pragma: no cover
    WebClient = None  # type: ignore
    _slack_import_error = exc
else:  # pragma: no cover - no error path when dependency present
    _slack_import_error = None


def _ensure_slack() -> None:
    if WebClient is None:  # pragma: no cover
        raise RuntimeError("slack-sdk is required for Slack notifications.") from _slack_import_error


def create_client(bot_token: str) -> "WebClient":
    _ensure_slack()
    return WebClient(token=bot_token)


def resolve_channel(policy_defaults: Any) -> str | None:
    return policy_defaults.slack_channel or os.getenv("G_GREMLIN_SLACK_CHANNEL")


def post_approval_request(
    web_client: Any,
    channel: str,
    plan: PlanV1,
    approve_token: str,
) -> Dict[str, Any]:
    blocks = build_plan_approval_blocks(plan, approve_token)
    text = f"Autopilot approval required for run {plan.run_id}"
    return web_client.chat_postMessage(channel=channel, text=text, **blocks)


def post_receipt_summary(
    web_client: Any,
    channel: str,
    receipt: ReceiptV1,
) -> Dict[str, Any]:
    changes_count = len(receipt.changes or [])
    blocks = [
        {
            "type": "header",
            "text": {"type": "plain_text", "text": "Autopilot Receipt"},
        },
        {
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": (
                    f"*Run:* {receipt.run_id}\n"
                    f"*Outcome:* {receipt.outcome}\n"
                    f"*Changes:* {changes_count}"
                ),
            },
        },
    ]
    text = f"Autopilot run {receipt.run_id} completed: {receipt.outcome}"
    return web_client.chat_postMessage(channel=channel, text=text, blocks=blocks)


__all__ = ["create_client", "resolve_channel", "post_approval_request", "post_receipt_summary"]



"""Offline Ed25519-signed license verification for g-gremlin."""

from __future__ import annotations

import base64
import hashlib
import json
import logging
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import Any

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey

log = logging.getLogger(__name__)

LICENSE_PATH = Path.home() / ".g_gremlin" / "license.json"
KEYRING_DIR = Path(__file__).resolve().parent / "keyring"


@dataclass
class LicensePayload:
    schema_version: int
    issuer: str
    kid: str
    license_id: str
    account_id: str
    plan: str
    trial: bool
    not_before: str
    issued_at: str
    expires_at: str
    grace_period_hours: int
    entitlements: dict[str, bool]
    limits: dict[str, int | None]
    entitlements_hash: str | None
    signature: str


@dataclass
class LicenseVerificationResult:
    valid: bool
    payload: LicensePayload | None
    error: str | None
    warnings: list[str]


def canonical_json_bytes(payload_dict: dict[str, Any]) -> bytes:
    """Return deterministic UTF-8 canonical JSON bytes for signing/verification."""
    return json.dumps(
        payload_dict,
        sort_keys=True,
        separators=(",", ":"),
        ensure_ascii=False,
        allow_nan=False,
    ).encode("utf-8")


def _parse_iso_utc(value: str) -> datetime:
    parsed = datetime.fromisoformat(value.replace("Z", "+00:00"))
    if parsed.tzinfo is None:
        return parsed.replace(tzinfo=timezone.utc)
    return parsed.astimezone(timezone.utc)


def _to_iso_utc(value: datetime) -> str:
    return value.astimezone(timezone.utc).replace(microsecond=0).isoformat().replace("+00:00", "Z")


def load_public_keyring(keyring_dir: Path | None = None) -> dict[str, Ed25519PublicKey]:
    directory = keyring_dir or KEYRING_DIR
    keys: dict[str, Ed25519PublicKey] = {}
    if not directory.exists():
        return keys

    for pem_path in sorted(directory.glob("*.pem")):
        kid = pem_path.stem
        try:
            key = serialization.load_pem_public_key(pem_path.read_bytes())
            if not isinstance(key, Ed25519PublicKey):
                log.warning("Skipping non-Ed25519 key file %s", pem_path)
                continue
            keys[kid] = key
        except Exception as exc:
            log.warning("Failed loading public key %s: %s", pem_path, exc)
    return keys


def compute_entitlements_hash(entitlements: dict[str, bool], limits: dict[str, int | None]) -> str:
    payload = {
        "entitlements": entitlements,
        "limits": limits,
    }
    digest = hashlib.sha256(canonical_json_bytes(payload)).hexdigest()
    return digest[:16]


def _parse_payload(raw: dict[str, Any]) -> LicensePayload:
    return LicensePayload(
        schema_version=int(raw.get("schema_version", 1)),
        issuer=str(raw.get("issuer") or ""),
        kid=str(raw.get("kid") or ""),
        license_id=str(raw.get("license_id") or ""),
        account_id=str(raw.get("account_id") or ""),
        plan=str(raw.get("plan") or "local_free"),
        trial=bool(raw.get("trial", False)),
        not_before=str(raw.get("not_before") or raw.get("issued_at") or ""),
        issued_at=str(raw.get("issued_at") or ""),
        expires_at=str(raw.get("expires_at") or ""),
        grace_period_hours=int(raw.get("grace_period_hours", 0) or 0),
        entitlements={
            str(k): bool(v)
            for k, v in (raw.get("entitlements") or {}).items()
        },
        limits={
            str(k): (None if v is None else int(v))
            for k, v in (raw.get("limits") or {}).items()
        },
        entitlements_hash=(str(raw.get("entitlements_hash")) if raw.get("entitlements_hash") is not None else None),
        signature=str(raw.get("signature") or ""),
    )


def verify_license(
    path: Path | None = None,
    keyring_dir: Path | None = None,
) -> LicenseVerificationResult:
    target_path = path or LICENSE_PATH
    if not target_path.exists():
        return LicenseVerificationResult(valid=False, payload=None, error="License file not found.", warnings=[])

    try:
        payload_raw = json.loads(target_path.read_text(encoding="utf-8"))
    except Exception as exc:
        return LicenseVerificationResult(valid=False, payload=None, error=f"Invalid license JSON: {exc}", warnings=[])

    if not isinstance(payload_raw, dict):
        return LicenseVerificationResult(valid=False, payload=None, error="License payload must be a JSON object.", warnings=[])

    signature_b64 = str(payload_raw.get("signature") or "")
    if not signature_b64:
        return LicenseVerificationResult(valid=False, payload=None, error="License signature missing.", warnings=[])

    payload_without_sig = dict(payload_raw)
    payload_without_sig.pop("signature", None)

    payload = _parse_payload(payload_raw)
    if not payload.kid:
        return LicenseVerificationResult(valid=False, payload=None, error="License key id (kid) missing.", warnings=[])

    keys = load_public_keyring(keyring_dir)
    signing_key = keys.get(payload.kid)
    if signing_key is None:
        return LicenseVerificationResult(
            valid=False,
            payload=None,
            error=f"Unknown license key id: {payload.kid}",
            warnings=[],
        )

    try:
        signature = base64.b64decode(signature_b64, validate=True)
    except Exception as exc:
        return LicenseVerificationResult(valid=False, payload=None, error=f"Invalid license signature encoding: {exc}", warnings=[])

    canonical = canonical_json_bytes(payload_without_sig)
    try:
        signing_key.verify(signature, canonical)
    except InvalidSignature:
        return LicenseVerificationResult(valid=False, payload=None, error="License signature verification failed.", warnings=[])

    warnings: list[str] = []
    expected_hash = compute_entitlements_hash(payload.entitlements, payload.limits)
    if payload.entitlements_hash and payload.entitlements_hash != expected_hash:
        warnings.append("entitlements_hash mismatch")

    return LicenseVerificationResult(valid=True, payload=payload, error=None, warnings=warnings)


def effective_entitlements(
    payload: LicensePayload,
    local_free_defaults: dict[str, bool],
    now_utc: datetime | None = None,
) -> dict[str, bool]:
    now = now_utc or datetime.now(timezone.utc)
    expires = _parse_iso_utc(payload.expires_at)
    grace_deadline = expires + timedelta(hours=payload.grace_period_hours)
    if now <= grace_deadline:
        return dict(payload.entitlements)
    return dict(local_free_defaults)


def effective_limits(
    payload: LicensePayload,
    local_free_limits: dict[str, int | None],
    now_utc: datetime | None = None,
) -> dict[str, int | None]:
    now = now_utc or datetime.now(timezone.utc)
    expires = _parse_iso_utc(payload.expires_at)
    grace_deadline = expires + timedelta(hours=payload.grace_period_hours)
    if now <= grace_deadline:
        return dict(payload.limits)
    return dict(local_free_limits)


def evaluate_license_window(payload: LicensePayload, now_utc: datetime | None = None) -> tuple[bool, bool, str | None]:
    """
    Returns (within_valid_window, within_grace, error).

    within_valid_window: now between not_before and expires_at inclusive.
    within_grace: now after expiry but within grace_period_hours.
    """
    now = now_utc or datetime.now(timezone.utc)

    try:
        not_before = _parse_iso_utc(payload.not_before)
        expires = _parse_iso_utc(payload.expires_at)
    except Exception as exc:
        return False, False, f"Invalid license time fields: {exc}"

    if now < not_before:
        return False, False, f"License not active until {_to_iso_utc(not_before)}"

    if now <= expires:
        return True, False, None

    grace_deadline = expires + timedelta(hours=payload.grace_period_hours)
    if now <= grace_deadline:
        return False, True, None

    return False, False, f"License expired at {_to_iso_utc(expires)}"
